/* 
 * txrxComp.c --
 *
 *    Contains the byte code generator for the XDR/RPC translator.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) txrxComp.c 1.1 96/08/13 15:18:56
 */

#include "txrxInt.h"           /* Internal defintions. */

/*
 * Local functions.
 */
static int              CodeInit _ANSI_ARGS_((Txrx_Tape *tapePtr));
static int              CodeRealloc _ANSI_ARGS_((Txrx_Tape *tapePtr));
static int              CodeClose _ANSI_ARGS_((Txrx_Tape *tapePtr));
static int              CodeSymbolBegin _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                     CellType type,
                                                     ClientData clientData));
static void             CodeSymbolEnd _ANSI_ARGS_((Txrx_Tape *tapePtr));
static void             CodeLocation _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                  SymbolData *symbolPtr));
static void             CodeValue _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                               ClientData clientData));
static int              CodeResolveName _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                     SymbolData *symbolPtr));
static int              CompileModule _ANSI_ARGS_((Txrx_Tape *tapePtr));
static int              CompileSymbol _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                   SymbolData *symbolPtr));
static int              CompileProgram _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    SymbolData *symbolPtr));
static int              CompileVersion _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    SymbolData *symbolPtr));
static int              CompileProcedure _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                      SymbolData *symbolPtr));
static int              CompileEnum _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                 SymbolData *symbolPtr));
static int              CompileStrAndOpq _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                      SymbolData *symbolPtr));
static int              CompileArray _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    SymbolData *symbolPtr));
static int              CompilePointer _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    SymbolData *symbolPtr));
static int              CompileStruct _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    SymbolData *symbolPtr));
static int              CompileUnion _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    SymbolData *symbolPtr));
static int              ResolveLabelValue _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                       Token *token,
                                                       SymbolData *symbol,
                                                       long *value));
static int              CheckDiscriminant _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                       SymbolData *symbolPtr,
                                                       SymbolData **dPtrPtr));

static int              PrintCell _ANSI_ARGS_((Tcl_Interp *interp,
                                               Tcl_DString *dsPtr,
                                               Txrx_Tape *tapePtr,
                                               int pc));

/*
 * Local variables.
 */

/*
 * Used to determine the state of building symbols. Is 1 after Code
 * SymbolBegin and before CodeSymbolEnd, 0 otherwise.
 */
int inSymbol = 0;


/*
 *----------------------------------------------------------------------
 *
 * Txrx_Compile --
 *
 *    Generate byte code from a symbol table.
 *
 * Results:
 *    A standard Tcl_Result.
 *
 * Side effects:
 *    Allocates data structures to store the tape.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_Compile(tapePtr)
     Txrx_Tape *tapePtr;
{
    int a;
    char *cPtr;

    /*
     * The translation to tape is also a recursive process.
     */
    if (CodeInit(tapePtr) == TCL_ERROR) {
        return TCL_ERROR;
    }
    
    if (CompileModule(tapePtr) == TCL_ERROR) {
        Tcl_DeleteHashTable(&tapePtr->patchTable);
        return TCL_ERROR;
    }

    if (CodeClose(tapePtr) == TCL_ERROR) {
        return TCL_ERROR;
    }
    tapePtr->endCode = tapePtr->codePC;
    if (tapePtr->endCode >= tapePtr->codeSize) {
        Tcl_AppendResult(tapePtr->interp, "end code is more than code size\n",
                         (char *)NULL);
        return TCL_ERROR;
    }
    if (strlen(tapePtr->interp->result) != 0) {
        return TCL_ERROR;
    }
    /*
     * Determine the endianness of the machine.
     */
    a = 0x11223344;
    cPtr = (char *)&a;
    tapePtr->endian = (*cPtr == 0x11)?TXRX_BIGENDIAN:TXRX_LITTLEENDIAN;
/*
    {
        char c0 = cPtr[0], c1 = cPtr[1], c2 = cPtr[2], c3 = cPtr[3];
        printf("%x %x %x %x \n", (int)c0, (int)c1, (int)c2, (int)c3);
        if (tapePtr->endian == TXRX_BIGENDIAN) {
            printf("bigEndian\n"); 
        } else {
            printf("littleEndian\n");
        }
    }
*/
       return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_PrintTape --
 *
 *    Prepare a list with the contents of the tape.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    interp->result is set to the description.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_PrintTape(interp, tapePtr)
     Tcl_Interp *interp;
     Txrx_Tape *tapePtr;
{
    SymbolData *symbolPtr;
    int crtPC = CODE_START;
    int oldPC;
    long cellType;
    Tcl_DString ds;            /* String where result accumulates.     */
    if (tapePtr == NULL || tapePtr->codePtr == NULL) {
        Tcl_AppendResult(interp, "bogus tape\n", (char *)NULL);
        return TCL_ERROR;
    }
    Tcl_DStringInit(&ds);
    while (crtPC < tapePtr->codePC) {
        if (tapePtr->codePtr[crtPC] != CELL_START) {
            Tcl_AppendResult(tapePtr->interp,
                             "cell start delimiter not found\n", (char *)NULL);
            return TCL_ERROR;
        }
        oldPC = crtPC;
        cellType = tapePtr->codePtr[crtPC + 1];
        symbolPtr = (SymbolData *)tapePtr->codePtr[crtPC + 2];
        crtPC = PrintCell(interp, &ds, tapePtr, crtPC);
        if (tapePtr->codePtr[crtPC - 1] != CELL_END) {
            Tcl_DStringFree(&ds);
            Tcl_AppendResult(tapePtr->interp, "cell end delimiter not found\n",
                             (char *)NULL);
            return TCL_ERROR;
        }
    }   
    Tcl_DStringResult(tapePtr->interp, &ds);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileModule --
 *
 *    Translate a symbol table (syntax tree) into a tape (linear code). 
 *
 * Results:
 *    A standard Tcl result..
 *
 * Side effects:
 *    Changes symbolPtr->codeAddress to a location on the tape.
 *
 *----------------------------------------------------------------------
 */

static int
CompileModule(tapePtr)
     Txrx_Tape *tapePtr;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    SymbolData *symbolPtr;
    int numPrograms = 0;
    /*
     * Compile all programs in the symbol table, then
     * assemble the module code.
     */
    entryPtr = Tcl_FirstHashEntry(&tapePtr->constTypeTable, &searchInfo);
    while (entryPtr) {
        symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (CompileSymbol(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (symbolPtr->type == SYM_PROGRAM) {
            numPrograms ++;
        }
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    /*
     * Now, assemble the code for the module.
     */
    tapePtr->entryPoint =
            CodeSymbolBegin(tapePtr, TMODULE, (ClientData) NULL);
    CodeValue(tapePtr, (ClientData)numPrograms);
    entryPtr = Tcl_FirstHashEntry(&tapePtr->constTypeTable, &searchInfo);
    while (entryPtr) {
        symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (symbolPtr->type == SYM_PROGRAM) {
            CodeValue(tapePtr, (ClientData)symbolPtr->constValue);
            CodeLocation(tapePtr, symbolPtr);
        }
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    CodeSymbolEnd(tapePtr);
    
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileSymbol --
 *
 *    Dispatch compile of symbol to appropriate function.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
CompileSymbol(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    if (symbolPtr->codeAddress != 0) {
        return TCL_OK;
    }
    switch (symbolPtr->type) {
      case SYM_PROGRAM:
        if (CompileProgram(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        break;
      case SYM_VERSION:
        if (CompileVersion(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        break;
      case SYM_PROCEDURE:
        if (CompileProcedure(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        break;
      case SYM_ENUM:
        if (CompileEnum(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        break;
      case SYM_STRING:
      case SYM_OPAQUE:
        if (CompileStrAndOpq(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        break;
      case SYM_ARRAY:
        if (CompileArray(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        break;
      case SYM_POINTER:
        if (CompilePointer(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        break;
      case SYM_STRUCT:
        if (CompileStruct(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        } 
        break;
      case SYM_UNION:
        if (CompileUnion(tapePtr, symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        } 
        break;
      case SYM_CONST:
        /* Constants eventually may compile to something. */
        break;

      case SYM_INT:
      case SYM_UNSIGNED:
      case SYM_BOOL:
      case SYM_HYPER:
      case SYM_UNSIGNEDHYPER:
      case SYM_FLOAT:
      case SYM_DOUBLE:
      case SYM_QUADRUPLE:
      case SYM_VOID:
        /*printf("symbol type %d has code address 0!\n", symbolPtr->type);*/
        break;
      default:
        /*
         * printf("Symbol %d has not compile function!\n", symbolPtr->type);
         */
        break;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileProgram --
 *
 *    Compile a program symbol into a linear bytecoded format.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Updates program symbol with tape location of the program.
 *
 *----------------------------------------------------------------------
 */

static int
CompileProgram(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    SymbolData *verSymbolPtr;
    int numVer = 0;
    /*
     * Compile all versions in the symbol table, then
     * assemble the program code.
     */
    entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
    while (entryPtr) {
        verSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (CompileSymbol(tapePtr, verSymbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (verSymbolPtr->type != SYM_VERSION) {
            Tcl_AppendResult(tapePtr->interp, "invalid symbol while ",
                             "compiling program\n", (char *)NULL);
            return TCL_ERROR;
        } 
        numVer ++;
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }

    /*
     * Code the program.
     */
    symbolPtr->codeAddress =
            CodeSymbolBegin(tapePtr, TPROGRAM, (ClientData) symbolPtr);
    CodeValue(tapePtr, (ClientData)numVer);
    CodeValue(tapePtr, (ClientData)symbolPtr->constValue);
    entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
    while (entryPtr) {
        verSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        CodeValue(tapePtr, (ClientData)verSymbolPtr->constValue);
        CodeLocation(tapePtr, verSymbolPtr);
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    CodeSymbolEnd(tapePtr);
    
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileVersion --
 *
 *    Compile a program version into a linear bytecoded format.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Updates version symbol with tape location of the program.
 *
 *----------------------------------------------------------------------
 */

static int
CompileVersion(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    SymbolData *procSymbolPtr;
    int numProcs = 0;
    /*
     * Compile all procedures in the version hash table.
     */
    entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
    while (entryPtr) {
        procSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (CompileSymbol(tapePtr, procSymbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (procSymbolPtr->type != SYM_PROCEDURE) {
            Tcl_AppendResult(tapePtr->interp, "invalid symbol while ",
                             "compiling version\n", (char *)NULL);
            return TCL_ERROR;
        } 
        numProcs ++;
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    
    /*
     * Code the program.
     */
    symbolPtr->codeAddress = 
            CodeSymbolBegin(tapePtr, TVERSION, (ClientData) symbolPtr);
    CodeValue(tapePtr, (ClientData)numProcs);
    CodeValue(tapePtr, (ClientData)symbolPtr->constValue);
    entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
    while (entryPtr) {
        procSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        CodeValue(tapePtr, (ClientData)procSymbolPtr->constValue);
        CodeLocation(tapePtr, procSymbolPtr);
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    CodeSymbolEnd(tapePtr);
    
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileProcedure --
 *
 *    Compile a procedure symbol into bytecode.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Updates procedure symbol with code location on the linear tape.
 *
 *----------------------------------------------------------------------
 */

static int
CompileProcedure(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Tcl_HashEntry *entryPtr;
    SymbolData *argSymbolPtr;
    int numArgs = (int)symbolPtr->identValue;
    int i;
    char buf[10];
    
    /*
     * Compile all arguments.
     */
    for (i = 0; i <= numArgs; i ++) {
        sprintf(buf, "%d", i);
        entryPtr = Tcl_FindHashEntry(symbolPtr->hashTablePtr, buf);
        argSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (CompileSymbol(tapePtr, argSymbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
    }
    
    /*
     * Code the program.
     */
    symbolPtr->codeAddress = 
            CodeSymbolBegin(tapePtr, TPROC, (ClientData) symbolPtr);
    CodeValue(tapePtr, (ClientData)numArgs);
    CodeValue(tapePtr, (ClientData)symbolPtr->constValue);

    for (i = 0; i <= numArgs; i ++) {
        sprintf(buf, "%d", i);
        entryPtr = Tcl_FindHashEntry(symbolPtr->hashTablePtr, buf);
        argSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        CodeLocation(tapePtr, argSymbolPtr);
    }
    CodeSymbolEnd(tapePtr);
    
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileEnum --
 *
 *    Compile an enumeration type to byecode.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Updates enum to hold the address on the tape where the
 *    bytecode for the enum is stored.
 *
 *----------------------------------------------------------------------
 */

static int
CompileEnum(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    int numValues = (int)symbolPtr->constValue;
    ClientData value;
    
    /*
     * Code the enumeration.
     */
    symbolPtr->codeAddress = 
            CodeSymbolBegin(tapePtr, TENUM, (ClientData) symbolPtr);
    CodeValue(tapePtr, (ClientData)numValues);
    entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
    while (entryPtr) {
        value = Tcl_GetHashValue(entryPtr);
        CodeValue(tapePtr, value);
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    CodeSymbolEnd(tapePtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileStrAndOpq --
 *
 *    Compile a string/opaque description to byecode.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Updates string/opaque symbol to hold address.
 *
 *----------------------------------------------------------------------
 */

static int
CompileStrAndOpq(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    int size = (int)symbolPtr->constValue;
    int isVariable = (int)symbolPtr->identValue;
    
    /*
     * Code.
     */
    symbolPtr->codeAddress = 
            CodeSymbolBegin(tapePtr, 
                            (symbolPtr->type == SYM_STRING)?TSTRING:TOPAQUE, 
                            (ClientData) symbolPtr);
    CodeValue(tapePtr, (ClientData)size);
    CodeValue(tapePtr, (ClientData)isVariable);
    CodeSymbolEnd(tapePtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileArray --
 *
 *    Translate to bytecode an array description.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Updates symbol to hold address.
 *
 *----------------------------------------------------------------------
 */

static int
CompileArray(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    int size = (int)symbolPtr->constValue;
    int isVariable = (int)symbolPtr->identValue;
    SymbolData *elmSymbolPtr = symbolPtr->referenceValue;

    /*
     * Code.
     */
    symbolPtr->codeAddress = 
            CodeSymbolBegin(tapePtr, TARRAY, (ClientData) symbolPtr);
    CodeValue(tapePtr, (ClientData)size);
    CodeValue(tapePtr, (ClientData)isVariable);
    CodeLocation(tapePtr, elmSymbolPtr);
    CodeSymbolEnd(tapePtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompilePointer --
 *
 *    Translate to bytecode a pointer symbol.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Updates symbol to hold address.
 *
 *----------------------------------------------------------------------
 */

static int
CompilePointer(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    SymbolData *elmSymbolPtr = symbolPtr->referenceValue;

    /*
     * Code.
     */
    symbolPtr->codeAddress = 
            CodeSymbolBegin(tapePtr, TPOINTER, (ClientData) symbolPtr);
    CodeLocation(tapePtr, elmSymbolPtr);
    CodeSymbolEnd(tapePtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileStruct --
 *
 *    Translate to bytecode a struct declaration.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Updates symbol to hold address on tape.
 *
 *----------------------------------------------------------------------
 */

static int
CompileStruct(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Tcl_HashEntry *entryPtr;
    SymbolData *fieldSymbolPtr;
    int numFields = (int)symbolPtr->constValue;
    char buf[10];
    int i;

    /*
     * Compile the fields.
     */
    for (i = 1; i <= numFields; i ++) {
        sprintf(buf, "%d", i);
        entryPtr = Tcl_FindHashEntry(symbolPtr->hashTablePtr, buf);
        fieldSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (CompileSymbol(tapePtr, fieldSymbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
    }
    /*
     * Code.
     */
    symbolPtr->codeAddress = 
            CodeSymbolBegin(tapePtr, TSTRUCT, (ClientData) symbolPtr);
    CodeValue(tapePtr, (ClientData)numFields);
    for (i = 1; i <= numFields; i ++) {
        sprintf(buf, "%d", i);
        entryPtr = Tcl_FindHashEntry(symbolPtr->hashTablePtr, buf);
        fieldSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        CodeLocation(tapePtr, fieldSymbolPtr);
    }
    CodeSymbolEnd(tapePtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CompileUnion --
 *
 *    Translate to bytecode a discriminated union declaration.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Updates symbol to hold address on tape.
 *
 *----------------------------------------------------------------------
 */

static int
CompileUnion(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    CaseData *defaultCasePtr = (CaseData *)symbolPtr->defaultValue;
    CaseData *casePtr;
    SymbolData *discrSymbolPtr = (SymbolData *)symbolPtr->referenceValue;
    SymbolData *discrPtr;
    int hasDefault = (int)symbolPtr->constValue;
    long value;
    int numCases;
    
    /*
     * The discriminant must be either an int or a bool or
     * an enum or a type reference to the former.
     */
    if (CheckDiscriminant(tapePtr, discrSymbolPtr, &discrPtr) == TCL_ERROR) {
        return TCL_ERROR;
    }

    if (CompileSymbol(tapePtr, discrSymbolPtr) == TCL_ERROR) {
        return TCL_ERROR;
    }
    /*
     * Compile the default.
     */
    if (   hasDefault 
        && CompileSymbol(tapePtr, defaultCasePtr->symbolPtr) == TCL_ERROR) {
        return TCL_ERROR;
    }
    
    numCases = 0;
    entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
    while (entryPtr) {
        numCases ++;
        casePtr = (CaseData *)Tcl_GetHashValue(entryPtr);
        if (CompileSymbol(tapePtr, casePtr->symbolPtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    /* 
     * Generate code.
     */
    symbolPtr->codeAddress = 
            CodeSymbolBegin(tapePtr, TUNION, (ClientData) symbolPtr);
    CodeLocation(tapePtr, discrSymbolPtr);
    CodeValue(tapePtr, (ClientData)numCases);
    CodeValue(tapePtr, (ClientData)hasDefault);
    if (hasDefault) {
        CodeLocation(tapePtr, defaultCasePtr->symbolPtr);
    } else {
        CodeValue(tapePtr, (ClientData)0);
    }
    entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
    while (entryPtr) {
        casePtr = (CaseData *)Tcl_GetHashValue(entryPtr);
        if (ResolveLabelValue(tapePtr, 
                              &casePtr->label, 
                              discrPtr,  
                              &value) == TCL_ERROR) {
            return TCL_ERROR;
        }
        CodeValue(tapePtr, (ClientData) value);
        CodeLocation(tapePtr, casePtr->symbolPtr);
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    CodeSymbolEnd(tapePtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ResolveLabelValue --
 *
 *    Determine the numeric value for a label of a case.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
ResolveLabelValue(tapePtr, token, symbolPtr, value)
     Txrx_Tape *tapePtr;
     Token *token;
     SymbolData *symbolPtr;
     long *value;
{
    Tcl_HashEntry *entryPtr; 
    SymbolData *symPtr;

    switch (symbolPtr->type) {
      case SYM_INT:
      case SYM_UNSIGNED:
        if (token->type != VALUE && token->type != IDENT) {
            Tcl_AppendResult(tapePtr->interp, 
                             "integer discriminant requires numeric labels\n",
                             (char *)NULL);
            return TCL_ERROR;
        }
        if (token->type == IDENT) {
            entryPtr = Tcl_FindHashEntry(&tapePtr->constTypeTable,
                                         token->identifier);
            if (entryPtr == NULL) {
                Tcl_AppendResult(tapePtr->interp, "label \"",
                                 token->identifier, "\" is not a valid ",
                                 "value for the discriminant\n", (char *)NULL);
                return TCL_ERROR;
            }
            symPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
            *value = symPtr->constValue;

        } else {
            if (symbolPtr->type == SYM_UNSIGNED && token->value < 0) {
                Tcl_AppendResult(tapePtr->interp, 
                                 "unsigned integer discriminant requires ",
                                 "positive numeric labels\n",
                                 (char *)NULL);
                return TCL_ERROR;
            }
            *value = (long)token->value;
        }
        break;
      case SYM_BOOL:
        if (token->type != VALUE) {
            if (token->type != IDENT) {
                goto errbool;
            }
            if (   strcmp(token->identifier, "false") == 0 
                || strcmp(token->identifier, "FALSE") == 0) {
                *value = 0;
            } else if (   strcmp(token->identifier, "true") == 0 
                       || strcmp(token->identifier, "TRUE") == 0) { 
                *value = 1;
            } else {
                goto errbool;
            }
            return TCL_OK;
        } 
        if (token->value != 0 && token->value != 1) {
          errbool:
            Tcl_AppendResult(tapePtr->interp, 
                             "boolean discriminant requires ",
                             "0 or 1 as labels\n",
                             (char *)NULL);
            return TCL_ERROR;
        }
        break;
      case SYM_ENUM:
        if (token->type != IDENT) {
            Tcl_AppendResult(tapePtr->interp, "label values must be ",
                             "identifiers defined in the discriminant enum\n",
                             (char *)NULL);
            return TCL_ERROR;
        }
        entryPtr = Tcl_FindHashEntry(symbolPtr->hashTablePtr, 
                                     token->identifier);
        if (entryPtr == NULL) {
            /*
             * Try it as a constant.
             */
            entryPtr = Tcl_FindHashEntry(&tapePtr->constTypeTable,
                                         token->identifier);
            if (entryPtr == NULL) {
                Tcl_AppendResult(tapePtr->interp, "label \"",
                                 token->identifier, "\" is not a valid ",
                                 "value for the discriminant\n", (char *)NULL);
                return TCL_ERROR;
            }
            symPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
            *value = symPtr->constValue;
        } else {
            *value = (long)Tcl_GetHashValue(entryPtr);
        }
        break;
      default:
        Tcl_AppendResult(tapePtr->interp, "invalid type in resolve\n",
                         (char *)NULL);
        return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CheckDiscriminant --
 *
 *    Check if a discriminant symbol is valid i.e. it is
 *    either an int or a bool or an enum.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Sets the parent symbol as result.
 *
 *----------------------------------------------------------------------
 */

static int
CheckDiscriminant(tapePtr, symbolPtr, parentPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     SymbolData **parentPtr;
{
    Tcl_HashEntry *entryPtr;
    int count = 0;

    while (count ++ < 100 && symbolPtr->type == SYM_TYPE) {
        /*
         * Get the parentSymbol from the type table.
         */
        entryPtr = Tcl_FindHashEntry(&tapePtr->constTypeTable,
                                     symbolPtr->identValue);
        if (entryPtr == NULL) {
            /* Cannot resolve type - should have been caught 
             * earlier by the compile.
             */
            Tcl_AppendResult(tapePtr->interp, "cannot find type \"",
                             symbolPtr->identValue, "\" in the symbol table\n",
                             (char *)NULL);
            return TCL_ERROR;
        }
        symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
    }
    if (count >= 100 || symbolPtr->type == SYM_TYPE) {
        Tcl_AppendResult(tapePtr->interp, "self referential type \"",
                         symbolPtr->identValue, "\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    
    /*
     * Check if allowed type.
     */
    switch (symbolPtr->type) {
      case SYM_INT:
      case SYM_BOOL:
      case SYM_UNSIGNED:
      case SYM_ENUM:
        *parentPtr = symbolPtr;
        break;
      default:
        Tcl_AppendResult(tapePtr->interp, "union discriminant type must ",
                         "be \"int\", \"unsigned int\", \"bool\" or ",
                         " an enumerated type\n", (char *)NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * PrintCell --
 *
 *    Append to interpeter result data about a bytecoded symbol.
 *
 * Results:
 *    Location on tape immediately after the end of the cell.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
PrintCell(interp, dsPtr, tapePtr, pc)
     Tcl_Interp *interp;
     Tcl_DString *dsPtr;
     Txrx_Tape *tapePtr;
     int pc;
{
    
    CellType typ = tapePtr->codePtr[pc + 1];
    char buf[256];
    int i;
    
    Tcl_DStringStartSublist(dsPtr);
    switch (typ) {
      case TMODULE:
        Tcl_DStringAppendElement(dsPtr,"MODULE");
        sprintf(buf, "%d", tapePtr->entryPoint);
        Tcl_DStringAppendElement(dsPtr, buf);
        sprintf(buf, "%d", tapePtr->endCode);
        Tcl_DStringAppendElement(dsPtr, buf);
        sprintf(buf, "%d", pc);
        Tcl_DStringAppendElement(dsPtr, buf);
        for (i = 0; i < tapePtr->codePtr[pc + 3]; i ++) {
            Tcl_DStringStartSublist(dsPtr);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 4 + 2 * i]);
            Tcl_DStringAppendElement(dsPtr, buf);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 5 + 2 * i]);
            Tcl_DStringAppendElement(dsPtr, buf);
            Tcl_DStringEndSublist(dsPtr);
        }
        pc += 5 + 2 * i;
        break;
      case TPROGRAM:
        Tcl_DStringAppendElement(dsPtr,"PROGRAM");
        sprintf(buf, "%ld", tapePtr->codePtr[pc + 4]);
        Tcl_DStringAppendElement(dsPtr, buf);
        sprintf(buf, "%d", pc);
        Tcl_DStringAppendElement(dsPtr, buf);
        for (i = 0; i < tapePtr->codePtr[pc + 3]; i ++) {
            Tcl_DStringStartSublist(dsPtr);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 5 + 2 * i]);
            Tcl_DStringAppendElement(dsPtr, buf);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 6 + 2 * i]);
            Tcl_DStringAppendElement(dsPtr, buf);
            Tcl_DStringEndSublist(dsPtr);
        }
        pc += 6 + 2 * i;
        break;
      case TVERSION:
        Tcl_DStringAppendElement(dsPtr,"VERSION");
        sprintf(buf, "%ld", tapePtr->codePtr[pc + 4]);
        Tcl_DStringAppendElement(dsPtr, buf);
        sprintf(buf, "%d", pc);
        Tcl_DStringAppendElement(dsPtr, buf);
        for (i = 0; i < tapePtr->codePtr[pc + 3]; i ++) {
            Tcl_DStringStartSublist(dsPtr);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 5 + 2 * i]);
            Tcl_DStringAppendElement(dsPtr, buf);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 6 + 2 * i]);
            Tcl_DStringAppendElement(dsPtr, buf);
            Tcl_DStringEndSublist(dsPtr);
        }
        pc += 6 + 2 * i;
        break;
      case TPROC:
        Tcl_DStringAppendElement(dsPtr,"PROC");
        sprintf(buf, "%ld", tapePtr->codePtr[pc + 4]);
        Tcl_DStringAppendElement(dsPtr, buf);
        sprintf(buf, "%d", pc);
        Tcl_DStringAppendElement(dsPtr, buf);
        for (i = 0; i <= tapePtr->codePtr[pc + 3]; i ++) {
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 5 + i]);
            Tcl_DStringAppendElement(dsPtr, buf);
        }
        pc += 6 + i;
        break;
      case TENUM:
        Tcl_DStringAppendElement(dsPtr,"ENUM");
        sprintf(buf, "%d", pc);
        Tcl_DStringAppendElement(dsPtr, buf);
        for (i = 0; i < tapePtr->codePtr[pc + 3]; i ++) {
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 4 + i]);
            Tcl_DStringAppendElement(dsPtr, buf);
        }
        pc += 5 + i;
        break;
      case TSTRING:
      case TOPAQUE:
        Tcl_DStringAppendElement(dsPtr, typ == TSTRING?"STRING":"OPAQUE");
        sprintf(buf, "%d", pc);
        Tcl_DStringAppendElement(dsPtr, buf);
        Tcl_DStringAppendElement(dsPtr,
                                 tapePtr->codePtr[pc + 4]?"FIXED":"VARIABLE");
        sprintf(buf, "%ld", tapePtr->codePtr[pc + 3]);
        Tcl_DStringAppendElement(dsPtr, buf);
        pc += 6;
        break;
      case TARRAY:
        Tcl_DStringAppendElement(dsPtr, "ARRAY");
        sprintf(buf, "%d", pc);
        Tcl_DStringAppendElement(dsPtr, buf);
        Tcl_DStringAppendElement(dsPtr,
                                 tapePtr->codePtr[pc + 4]?"FIXED":"VARIABLE");
        sprintf(buf, "%ld", tapePtr->codePtr[pc + 3]);
        Tcl_DStringAppendElement(dsPtr, buf);
        sprintf(buf, "%ld", tapePtr->codePtr[pc + 5]);
        Tcl_DStringAppendElement(dsPtr, buf);
        pc += 7;
        break;
      case TPOINTER:
        Tcl_DStringAppendElement(dsPtr, "POINTER");
        sprintf(buf, "%d", pc);
        Tcl_DStringAppendElement(dsPtr, buf);
        sprintf(buf, "%ld", tapePtr->codePtr[pc + 3]);
        Tcl_DStringAppendElement(dsPtr, buf);
        pc += 5;
        break;
      case TSTRUCT:
        Tcl_DStringAppendElement(dsPtr, "STRUCT");
        sprintf(buf, "%ld", tapePtr->codePtr[pc + 3]);
        Tcl_DStringAppendElement(dsPtr, buf);
        sprintf(buf, "%d", pc);
        Tcl_DStringAppendElement(dsPtr, buf);
        for (i = 0; i < tapePtr->codePtr[pc + 3]; i ++) {
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 4 + i]);
            Tcl_DStringAppendElement(dsPtr, buf);
        }
        pc += 5 + i;
        break;
      case TUNION:
        if (tapePtr->codePtr[pc + 5] == 0) {
            /* 
             * No default.
             */
            Tcl_DStringAppendElement(dsPtr, "UNION");
            sprintf(buf, "%d", pc);
            Tcl_DStringAppendElement(dsPtr, buf);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 3]);
            Tcl_DStringAppendElement(dsPtr, buf);
        } else {
            /* 
             * Has default.
             */
            Tcl_DStringAppendElement(dsPtr, "UNIOND");
            sprintf(buf, "%d", pc);
            Tcl_DStringAppendElement(dsPtr, buf);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 3]);
            Tcl_DStringAppendElement(dsPtr, buf);
            Tcl_DStringStartSublist(dsPtr);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 6]);
            Tcl_DStringAppendElement(dsPtr, buf);
            Tcl_DStringEndSublist(dsPtr);
        }
        for (i = 0; i < tapePtr->codePtr[pc + 4]; i ++) {
            Tcl_DStringStartSublist(dsPtr);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 7 + 2 * i]);
            Tcl_DStringAppendElement(dsPtr, buf);
            sprintf(buf, "%ld", tapePtr->codePtr[pc + 8 + 2 * i]);
            Tcl_DStringAppendElement(dsPtr, buf);
            Tcl_DStringEndSublist(dsPtr);
        }
        pc += 8 + 2 * i;
        break;
      default:
        Tcl_DStringAppendElement(dsPtr,"unknown cell type");
        break;
    }
    Tcl_DStringEndSublist(dsPtr);
    return pc;
}

/*
 *----------------------------------------------------------------------
 *
 * CodeInit --
 *
 *    Initialize a byte code data structure.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Allocates some memory.
 *
 *----------------------------------------------------------------------
 */

static int
CodeInit(tapePtr)
     Txrx_Tape *tapePtr;
{
    tapePtr->codePtr = (long *)ckalloc(CODE_INCREMENT * sizeof(long));
    if (tapePtr->codePtr == NULL) {
        Tcl_AppendResult(tapePtr->interp, "cannot allocate memory for tape\n",
                         (char *)NULL);
        return TCL_ERROR;
    }
    tapePtr->codePC = CODE_START;
    tapePtr->codeSize = CODE_INCREMENT;

    /*
     * Initialize the code patch table that stores locations and
     * pointers to the type table entries representing the type.
     */
    Tcl_InitHashTable(&tapePtr->patchTable, TCL_ONE_WORD_KEYS);
    inSymbol = 0;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CodeRealloc --
 *
 *    Expand a byte code data structure.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Reallocates some memory.
 *
 * Note:
 *    It is not supposed to be called outside of the Codexxx procedures.
 *----------------------------------------------------------------------
 */

static int
CodeRealloc(tapePtr)
     Txrx_Tape *tapePtr;
{
    long *newCode;
    newCode = (long *)ckrealloc((char *)tapePtr->codePtr,
                                (tapePtr->codeSize + CODE_INCREMENT)
                                * sizeof(long));
    if (newCode == NULL) {
        Tcl_AppendResult(tapePtr->interp, "cannot reallocate code\n",
                         (char *) NULL);
        /*printf("out of memory %ld\n", (long) tapePtr->codeSize);*/
        return TCL_ERROR;
    }
    tapePtr->codeSize += CODE_INCREMENT;
    tapePtr->codePtr = newCode;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CodeClose --
 *
 *    Terminate building a fragment of code.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
CodeClose(tapePtr)
     Txrx_Tape *tapePtr;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    SymbolData *symbolPtr;
    long addr;

    entryPtr = Tcl_FirstHashEntry(&tapePtr->patchTable, &searchInfo);
    while (entryPtr) {
        symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        addr = (long) Tcl_GetHashKey(&tapePtr->patchTable, entryPtr);

        if (symbolPtr->codeAddress == 0) {
            /*printf("can't patch undefined symbol\n");*/
            Tcl_AppendResult(tapePtr->interp, "can't patch undefined symbol\n",
                             (char *)NULL);
            Tcl_DeleteHashTable(&tapePtr->patchTable);
            return TCL_ERROR;
        } 
        if (tapePtr->codePtr[addr] != 0) {
            /*printf("invalid code address %ld - contents was %ld\n",
                   addr, tapePtr->codePtr[addr]);*/
            Tcl_AppendResult(tapePtr->interp, "invalid code while patching\n",
                             (char *)NULL);
            Tcl_DeleteHashTable(&tapePtr->patchTable);
            return TCL_ERROR;
        }
        tapePtr->codePtr[addr] = symbolPtr->codeAddress;
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    Tcl_DeleteHashTable(&tapePtr->patchTable);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CodeSymbolBegin --
 *
 *    Start coding a symbol.
 *
 * Results:
 *    Offset on tape where symbol begins. Returns 0 in case of error. 
 *
 * Side effects:
 *    Changes tape.
 *
 *----------------------------------------------------------------------
 */

static int
CodeSymbolBegin(tapePtr, typ, clientData)
     Txrx_Tape *tapePtr;
     CellType typ;
     ClientData clientData;
{
    int res = tapePtr->codePC;
    int pc = tapePtr->codePC;
    if (inSymbol) {
        /*printf("in symbol\n");*/
        Tcl_AppendResult(tapePtr->interp, "in symbol\n", (char *)NULL);
        return 0;
    }
    if (pc + 2 >= tapePtr->codeSize) {
        CodeRealloc(tapePtr);
    }
    tapePtr->codePtr[pc++] = CELL_START;
    tapePtr->codePtr[pc++] = (long)typ;
    tapePtr->codePtr[pc++] = (long)clientData;
    inSymbol = 1;
    tapePtr->codePC = pc;
    return (int)res;
}

/*
 *----------------------------------------------------------------------
 *
 * CodeSymbolEnd --
 *
 *    Terminate building a fragment of code.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static void
CodeSymbolEnd(tapePtr)
     Txrx_Tape *tapePtr;
{
    if (!inSymbol) {
        /*printf("not in symbol0\n");*/
        Tcl_AppendResult(tapePtr->interp, "not in symbol0\n", (char *)NULL);
        return;
    }
    if (tapePtr->codePC >= tapePtr->codeSize) {
        CodeRealloc(tapePtr);
    }
    tapePtr->codePtr[tapePtr->codePC++] = CELL_END;
    inSymbol = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * CodeLocation --
 *
 *    Add a tape offset to the tape. Check if the tape offset
 *    points to a proper symbol.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Modifies code.
 *
 *----------------------------------------------------------------------
 */

static void
CodeLocation(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    int codeLocation;

    if (symbolPtr == NULL) {
        Tcl_AppendResult(tapePtr->interp, "cannot generate code location",
                         " for NULL symbol\n", (char *)NULL);
        return;
    }
    codeLocation = symbolPtr->codeAddress;

    if (!inSymbol) {
        /*printf("not in symbol1\n");*/
        Tcl_AppendResult(tapePtr->interp, "not in symbol1\n", (char *)NULL);
        return;
    }
    /* 
     * Sometimes the code address for a symbol might be 0 i.e. not yet
     * known. This should only happen for SYM_TYPE symbols which
     * are references to other symbols and that usually means the
     * type is not compiled.
     * 
     * The solution is to allocate the location and store into 
     * a table the number of the location and the name of the 
     * type. 
     *
     * When the compile is finished and all types have been 
     * allocated the code is patched with the acual code addresses
     * of the types. 
     * 
     * This is required by recursive type definitions and mutually
     * recursive data structures.
     * 
     * It is an error if after Compile a top level type has no 
     * address allocated to it - this means the type has not been
     * defined.
     * 
     * It is also an error if at CodeLocation time a type name
     * is not found in the type table.
     */

    if (tapePtr->codePC >= tapePtr->codeSize) {
        CodeRealloc(tapePtr);
    }

    if (symbolPtr->type != SYM_TYPE) {
        /* 
         * This is not a type reference. Proceed as required.
         */
        if (codeLocation == 0) {
            Tcl_AppendResult(tapePtr->interp, "internal error: symbol is ",
                             " not allocated\n", (char *)NULL);
            /*printf("symbol type %d is not allocated\n", symbolPtr->type);*/
        }
        
        if (codeLocation > CODE_START) {
            if (tapePtr->codePtr[codeLocation] != CELL_START) {
                Tcl_AppendResult(tapePtr->interp, "invalid location\n",
                                 (char *)NULL);
            }
        }
    } else {
        /* Store some value on tape. The value will indicate 
         * that a patch is needed at a later time. If the type 
         * is found and has already been allocated, then 
         * store the address of the type. Otherwise store whatever 
         * Resolve returns.
         */
        codeLocation = CodeResolveName(tapePtr, symbolPtr);
    }
    tapePtr->codePtr[tapePtr->codePC++] = codeLocation;
        
        
}

/*
 *----------------------------------------------------------------------
 *
 * CodeValue --
 *
 *    Add a numeric value to the tape. 
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Modifies tape code.
 *
 *----------------------------------------------------------------------
 */

static void
CodeValue(tapePtr, clientData)
     Txrx_Tape *tapePtr;
     ClientData clientData;
{
    if (tapePtr->codePC >= tapePtr->codeSize) {
        CodeRealloc(tapePtr);
    }

    if (!inSymbol) {
        /*printf("not in symbol2\n");*/
        Tcl_AppendResult(tapePtr->interp, "not in symbol2\n", (char *)NULL);
        return;
    }
    
    if (tapePtr->codePC >= tapePtr->codeSize) {
        CodeRealloc(tapePtr);
    }
    tapePtr->codePtr[tapePtr->codePC++] = (long)clientData;
}
/*
 *----------------------------------------------------------------------
 *
 * CodeResolveName --
 *
 *    Resolve the addresss of a type symbol.
 *
 * Results:
 *    Location of the type on tape or 0 to indicate the type
 *    is not yet compiled, i.e. translated onto tape, and -1
 *    signifying an error has occured.
 *
 * Side effects:
 *    If the type is allocated, changes the code address for the
 *    symbol passed as argument, otherwise it stores the type
 *    name and the current tapePC into a patchTable.
 *
 *----------------------------------------------------------------------
 */

static int
CodeResolveName(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Tcl_HashEntry *entryPtr;
    SymbolData *parentSymbolPtr;
    int isNew = 0;
    int loopCount = 0;

    if (symbolPtr->codeAddress != 0) {
        /*
         * The symbol already has an address, return it
         * as result.
         */
        return symbolPtr->codeAddress;
    }
    while (loopCount < 100 && symbolPtr->type == SYM_TYPE) {
        loopCount ++;
        if (   symbolPtr->identValue == NULL 
            || symbolPtr->identValue[0] == '\0'){
            Tcl_AppendResult(tapePtr->interp, 
                             "type name value is null or empty", (char *)NULL);
            return -1;
        }
        entryPtr = Tcl_FindHashEntry(&tapePtr->constTypeTable,
                                     symbolPtr->identValue);
        if (entryPtr == NULL) {
            /*
             * Type does not exist in the type table. 
             */
            Tcl_AppendResult(tapePtr->interp, "type \"", 
                             symbolPtr->identValue, "\" is not defined\n",
                             (char *)NULL);
            return -1;
        }
        parentSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (parentSymbolPtr->type != SYM_TYPE) {
            /*
             * The real thing - this is a data type definition proper.
             */
            if (parentSymbolPtr->codeAddress != 0) {
                /*
                 * Data type has already been defined, store
                 * the address into the type symbol and return
                 * it.
                 */
                symbolPtr->codeAddress = parentSymbolPtr->codeAddress;
                return symbolPtr->codeAddress;
            } else {
                /*
                 * Data type is not yet compiled to tape. Store 
                 * the address of the code and the symbol 
                 * into the patchTable table.
                 */
                entryPtr = Tcl_CreateHashEntry(&tapePtr->patchTable, 
                                               (char *)tapePtr->codePC, 
                                               &isNew);
                if (isNew == 0) {
                    /*printf("error while adding address to patch table\n");*/
                }
                Tcl_SetHashValue(entryPtr, (ClientData) parentSymbolPtr); 
                return 0;
            }
        }
        /*
         * The "refered to" (parent) symbol is yet another type symbol.
         * Change symbolPtr to point to it and loop.
         */
        symbolPtr = parentSymbolPtr;
    }
    if (symbolPtr->type == SYM_TYPE) {
        Tcl_AppendResult(tapePtr->interp, "self referencing type \"",
                         symbolPtr->identValue, "\"\n", (char *)NULL);
        return -1;   
    }
    Tcl_AppendResult(tapePtr->interp, "invalid symbol as argument\n", 
                     (char *)NULL);

    return -1;
}

/*
 *----------------------------------------------------------------------
 *
 * <procName> --
 *
 *    <description>
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static void
Proc()
{
}
