/* 
 * txrxParser.c --
 *
 *    Contains the translator from the XDR/RPC language to
 *    bytecode format. 
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) txrxPars.c 1.1 96/08/13 15:18:59
 */

#include "txrxInt.h"           /* Internal defintions. */

/*
 * Local functions.
 */
static int              Parse _ANSI_ARGS_((Txrx_Tape *tapePtr));
static int              GetChar _ANSI_ARGS_((Txrx_Tape *tapePtr, char *buf));
static void             UngetChar _ANSI_ARGS_((Txrx_Tape *tapePtr, int c));
static int              MatchKeyword _ANSI_ARGS_((char *buf, int len));
static int              Scan _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                          Token *token, 
                                          int tokenType));
static void             UnScan _ANSI_ARGS_((Txrx_Tape *tapePtr, Token *token));
static void             AddLineErrorInfo _ANSI_ARGS_((Txrx_Tape *tapePtr));

static int              Specification _ANSI_ARGS_((Txrx_Tape *tapePtr));

static int              ProgramDef _ANSI_ARGS_((Txrx_Tape *tapePtr));
static int              NumberInTable _ANSI_ARGS_((long version,
                                                   Tcl_HashTable *tablePtr,
                                                   int symbolType));
static int              VersionDef _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                SymbolData *symbolPtr,
                                                char *name));
static int              ProcedureDef _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                  SymbolData *symbolPtr,
                                                  char *name));
static int              TypeDef _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                             Token *token));
static int              ConstantDef _ANSI_ARGS_((Txrx_Tape *tapePtr));
static int              EnumBody _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                              SymbolData *symbolPtr));
static int              StructBody _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                SymbolData *symbolPtr));
static int              UnionBody _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                               SymbolData *symbolPtr));
static int              Declaration _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                 SymbolData *symbolPtr,
                                                 char *typeName));
static int              NamedType _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                               SymbolData *symbolPtr,
                                               Token *token,
                                               char *typeName));
static int              OpaqueDeclaration _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                       SymbolData *symbolPtr,
                                                       char *typeName));
static int              StringDeclaration _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                       SymbolData *symbolPtr,
                                                       char *typeName));
static int              TypeSpecifier _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                   SymbolData *symbolPtr,
                                                   Token *token));
static int              GetArraySize _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                  Token *token, 
                                                  long *value,
                                                  int *isVar));
static int              GetValue _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                              Token *token, long *value));
static int              Discriminant _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                  SymbolData *symbolPtr,
                                                  char *namePtr));
static int              Case _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                          SymbolData *symbolPtr,
                                          char *namePtr,
                                          int *isDefault,
                                          Token *labelPtr));
static int              AddToTable _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                Tcl_HashTable *tablePtr,
                                                char *id,
                                                ClientData clientData));

static void             FreeSymbol _ANSI_ARGS_((SymbolData *symbolPtr));
static void             FreeCase _ANSI_ARGS_((CaseData *caseDataPtr));
static void             FreeSymbolTable _ANSI_ARGS_((Tcl_HashTable *tablePtr));
static void             FreeCaseTable _ANSI_ARGS_((Tcl_HashTable *tablePtr));

static int              PrintSymbol _ANSI_ARGS_((Txrx_Tape *tape,
                                                 Tcl_DString *dsPtr,
                                                 SymbolData *symbolPtr,
                                                 char *name));
static SymbolData*      InitSymbol _ANSI_ARGS_(());
static CaseData*        InitCase _ANSI_ARGS_((SymbolData *symbolPtr,
                                              Token *label,
                                              char *name));
                                          

/*
 *----------------------------------------------------------------------
 *
 * Txrx_CreateTape --
 *
 *    Parse a tape description.
 *
 * Results:
 *    Pointer to tape data structure, NULL if the protocol cannot be parsed.
 *
 * Side effects:
 *    Allocates data structures needed to store protocol description.
 *
 *----------------------------------------------------------------------
 */

Txrx_Tape*
Txrx_CreateTape(interp, inText)
     Tcl_Interp* interp;
     char *inText;
{
    Txrx_Tape *tapePtr;           /* Tape being initialized. */
    
    tapePtr = (Txrx_Tape *)ckalloc(sizeof(Txrx_Tape));
    tapePtr->command = NULL;
    tapePtr->handlePtr = NULL;
    tapePtr->interp = interp;
    tapePtr->codePtr = NULL;

    tapePtr->dataOp = NONE;
    tapePtr->dataPtr = NULL;
    tapePtr->dataSize = 0;
    tapePtr->dataPos = 0;
    
    tapePtr->endian = TXRX_BIGENDIAN; /* In honor of the SUN team. */
    /*
     * Initialize symbol table.
     */
    Tcl_InitHashTable(&tapePtr->constTypeTable, TCL_STRING_KEYS);
    
    /*
     * Initialize scanner state for parsing predefined symbols.
     * Parse predefined symbols - rpc header and authorization
     * data structures.
     */
    tapePtr->inText = RPCV2DEF;
    tapePtr->inPos = 0;
    tapePtr->inputAvailable = 0;
    tapePtr->inputChar = 0;
    tapePtr->lexemeAvailable = 0;
    if (Parse(tapePtr) == TCL_ERROR) {
        Tcl_AppendResult(interp, "internal error while parsing definitions",
                         (char *)NULL);
        /*printf("error %s\n", tapePtr->interp->result);*/
        goto error;
    }
    /*
     * Initialize scanner state.
     */
    tapePtr->inText = inText;
    tapePtr->inPos = 0;
    tapePtr->inLine = 0;
    tapePtr->inChar = 0;
    tapePtr->inputAvailable = 0;
    tapePtr->inputChar = 0;
    tapePtr->lexemeAvailable = 0;
    
    if (Parse(tapePtr) == TCL_ERROR) {
        goto error;
    }
    if (Txrx_Compile(tapePtr) == TCL_ERROR) {
        goto error;
    }
    
    return tapePtr;

  error:
    /*
     * Compile did not succeed, free all symbols.
     */
    if (tapePtr->codePtr != NULL) {
        ckfree((char *)tapePtr->codePtr);
    }
    FreeSymbolTable(&tapePtr->constTypeTable);
    ckfree((char *)tapePtr);
    return NULL; 
}

/*
 *----------------------------------------------------------------------
 *
 * DeleteTape --
 *
 *    Free a tape description.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Deallocates data structure.
 *
 *----------------------------------------------------------------------
 */

void
Txrx_DeleteTape(tapePtr)
     Txrx_Tape *tapePtr;
{

    /*
     * If there is a data buffer allocated, free it.
     */
    
    if (tapePtr->codePtr != NULL) {
        ckfree((char *)tapePtr->codePtr);
    }
    if (tapePtr->dataPtr != NULL) {
        ckfree((char *)tapePtr->dataPtr);
    }
    FreeSymbolTable(&tapePtr->constTypeTable);
    ckfree((char *)tapePtr);
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_PrintSymbols --
 *
 *    Prepare a list with the description of the protocol.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    interp->result is set to the description.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_PrintSymbols(tapePtr)
     Txrx_Tape *tapePtr;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    SymbolData *symbolPtr;
    Tcl_DString ds;                 /* String used to store the printed
                                     * symbol list.
                                     */

    /*
     * Print each symbol in the type table.
     */
    entryPtr = Tcl_FirstHashEntry(&tapePtr->constTypeTable, &searchInfo);
    Tcl_DStringInit(&ds);
    while (entryPtr) {
        symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        PrintSymbol(tapePtr, &ds,
                    symbolPtr, 
                    Tcl_GetHashKey(&tapePtr->constTypeTable, entryPtr));
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    } 
    Tcl_DStringResult(tapePtr->interp, &ds);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Scan --
 *
 *    Read from a channel  - usually a text file -  and lexically
 *      analyze the input.
 *
 * Results:
 *    TCL_OK if input was lexically correct.
 *
 * Side effects:
 *    Out parameter 'token' contains a lexical symbol.
 *
 *----------------------------------------------------------------------
 */

static int
Scan(tapePtr, token, tokenType)
     Txrx_Tape *tapePtr;
     Token* token;
     int tokenType;
{
    int i = 0;
    int skip = 1;
    char buf[2];
    
    buf[1] = '\0';
    
    /* Check if a symbol isn't already available. */
    if (tapePtr->lexemeAvailable) {
        tapePtr->lexemeAvailable = 0;
        *token = tapePtr->lexeme;
        if (tokenType != UNKNOWN && tokenType != token->type) {
            Tcl_AppendResult(tapePtr->interp, "unexpected symbol as input\n",
                             (char *)NULL);
            return TCL_ERROR;
        }
        return TCL_OK;
    }
    
    do {
        /* Skip white space and comments. */
        while (GetChar(tapePtr, buf) && isspace(buf[0]));
        if (buf[0] == '\0') { 
            token->type = EOFFOUND;
            if (tokenType != UNKNOWN) {
                Tcl_AppendResult(tapePtr->interp, 
                                 "end of file while scanning\n", 
                                 (char *)NULL);
                return TCL_ERROR;
            }
            return TCL_OK;
        }
        
        /* buf[0] is possibly the first character of a token, identify. */
        if (buf[0] == '/') { 
            /* XDR specs "/" as the beginning of a comment. */
            buf[0] = tapePtr->inText[tapePtr->inPos ++];
            if (tapePtr->inText[tapePtr->inPos] == '\0') {
                Tcl_AppendResult(tapePtr->interp, 
                                 "End of file in middle of comment.",
                                 (char *)NULL);
                return TCL_ERROR;
            }
            if (buf[0] == '*') { /* A standard C comment was found. */
                do {
                    /* Skip over all characters, except for '*'. */
                    while (   (buf[0] = tapePtr->inText[tapePtr->inPos ++]) 
                           && buf[0] != '*');
                    /* When a '*' is found 
                     * check for the next character. A '/'
                     * signifies end of comment.
                     */
                } while (   (buf[0] = tapePtr->inText[tapePtr->inPos ++]) 
                         && buf[0] != '/');
                if (buf[0] != '/') {
                    Tcl_AppendResult(tapePtr->interp, 
                                     "End of file in middle of a comment.",
                                     (char *)NULL);
                    return TCL_ERROR;
                }
            } else if (buf[0] == '/') {
                /* A non standard C++ comment - skip to end of line.
                 */
                while (   (buf[0] = tapePtr->inText[tapePtr->inPos ++]) 
                       && buf[0] != '\n');
            }
        } else if (buf[0] == '#') {
            /* Tcl  - or C preprocessor directive -  comment.
             */
            while (   (buf[0] = tapePtr->inText[tapePtr->inPos ++]) 
                   && buf[0] != '\n');
        } else if (buf[0] == '%') {
            /* RPCGEN pass-through comment.
             */
            while (   (buf[0] = tapePtr->inText[tapePtr->inPos ++]) 
                   && buf[0] != '\n');
        } else {
            /* Not a comment, must be a token relevant to the parser. */
            skip = 0;
        }
    } while (skip);
    
    /* buf[0] is the first character of a token. */
    if (isalpha(buf[0]) || buf[0] == '_') {
        /* Identifier. */
        token->type = IDENT;
        do {
            if (i < MAXIDENTLEN) {
                token->identifier[i++] = buf[0];
            }
        } while (   (buf[0] = tapePtr->inText[tapePtr->inPos ++]) 
                 && (isalnum(buf[0]) || buf[0] == '_'));
        token->identifier[i] = '\0';
        token->length = i;
        UngetChar(tapePtr, buf[0]);
        /* The identifier may be a keyword. */
        if (MatchKeyword(token->identifier, token->length)) {
            token->type = KEYWRD;
            if (tokenType != UNKNOWN && tokenType != KEYWRD) {
                Tcl_AppendResult(tapePtr->interp, 
                                 "scanned unexpected keyword\n", 
                                 (char *)NULL);
                return TCL_ERROR;
            }
            return TCL_OK;
        }
        if (tokenType != UNKNOWN && tokenType != IDENT) {
            Tcl_AppendResult(tapePtr->interp, 
                             "scanned unexpected identifier\n",
                             (char *)NULL);
            return TCL_ERROR;
        }
        return TCL_OK;
    }
    
    if (isdigit(buf[0]) || buf[0] == '-') {
        int sign = 1;
        int base = 10;
        
        /* Scanning a constant. */
        if (tokenType != UNKNOWN && tokenType != VALUE) {
            Tcl_AppendResult(tapePtr->interp, 
                             "found unexpected numeric constant\n",
                             (char *)NULL);
            return TCL_ERROR;
        }
        
        token->type = VALUE;
        token->value = 0.0;
        if (buf[0] == '-') {/* Skip all white space up to the next digit. */
            sign = -1;
            while (   (buf[0] = tapePtr->inText[tapePtr->inPos ++]) 
                   && isspace(buf[0]));
        }
        if (!isdigit(buf[0])) {
            Tcl_AppendResult(tapePtr->interp, 
                             "Minus sign not followed by a number.",
                             (char *)NULL);
            return TCL_ERROR;
        }
        /* Numeric constant. */
        if (buf[0] == '0') { /* Either a hex or an octal constant. */
            if (!(buf[0] = tapePtr->inText[tapePtr->inPos ++])) {
                /* Read error or EOF - return as zero value and let the next
                 * step deal with the problem.
                 */
                return TCL_OK;
            }
            if (buf[0] == 'x' || buf[0] == 'X') {
                /* Got a hex constant. */
                base = 16;
                if (   !(buf[0] = tapePtr->inText[tapePtr->inPos ++]) 
                    || !isxdigit(buf[0])) {
                    Tcl_AppendResult(tapePtr->interp,
                                     "Invalid character",
                                     " in hexadecimal constant.",
                                     (char *)NULL);
                    return TCL_ERROR;
                }
            } else if (isdigit(buf[0])) {
                /* Got an octal constant. */
                base = 8;
            } else {
                /* Got a zero, put the character back. */
                UngetChar(tapePtr, buf[0]);
                return TCL_OK;
            }
        }
        do {
            if (i > MAXVALUELEN) {
                Tcl_AppendResult(tapePtr->interp, 
                                 "Too many characters in constant.",
                                 (char *)NULL);
                return TCL_ERROR;
            }
            i ++;
            token->value = 
                    token->value * (double)base + (double)(buf[0] - '0');
        } while (   (buf[0] = tapePtr->inText[tapePtr->inPos ++])
                 && ((base == 10 && isdigit(buf[0]))
                     || (base == 8  && isdigit(buf[0]) && buf[0] < '9')
                     || (base == 16  && isxdigit(buf[0]))));
        token->value *= (double)sign;
        UngetChar(tapePtr, buf[0]);
        return TCL_OK;
    }
    /* Special symbol, or just garbage. */
    
    switch (buf[0]) { /* Do not use to represent as token symbol the token
                       * is may not be portable.
                       */
      case '*':
        token->type = STAR;
        break;
      case '(':
        token->type = LPAREN;
        break;
      case ')':
        token->type = RPAREN;
        break;
      case '[':
        token->type = LSQUARE;
        break;
      case ']':
        token->type = RSQUARE;
        break;
      case '{':
        token->type = LCURLY;
        break;
      case '}':
        token->type = RCURLY;
        break;
      case '<':
        token->type = LANGLE;
        break;
      case '>':
        token->type = RANGLE;
        break;
      case ':':
        token->type = COLUMN;
        break;
      case ';':
        token->type = SEMICOLUMN;
        break;
      case '=':
        token->type = EQUAL;
        break;
      case '-':
        token->type = MINUS;
        break;
      case ',':
        token->type = COMMA;
        break;
      default:
        token->type = UNKNOWN;
        token->identifier[0] = buf[0];
        token->identifier[1] = '\0';
        token->length = 1;
        if (token->type == UNKNOWN) {
            Tcl_AppendResult(tapePtr->interp, 
                             "scanned unexpected character \"", 
                             token->identifier, "\"\n", (char *)NULL);
            return TCL_ERROR;
        }
        break;
    }
    if (tokenType != UNKNOWN && tokenType != token->type) {
        Tcl_AppendResult(tapePtr->interp, 
                         "scanned unexpected special character \"",
                         buf, "\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * UnScan --
 *
 *    Put back in the input stream a lexical atom.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Variable 'lexeme' contains the token for the next call to Scan.
 *
 *----------------------------------------------------------------------
 */

static void
UnScan(tapePtr, token)
     Txrx_Tape *tapePtr;
     Token *token;
{
    tapePtr->lexeme = *token;
    tapePtr->lexemeAvailable = 1;
}

/*
 *----------------------------------------------------------------------
 *
 * GetChar --
 *
 *    Read a character from a channel. May read back a character
 *      put back in the input stream by UngetChar.
 *
 * Results:
 *    Whatever Tcl_Chan returns.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
GetChar(tapePtr, buf)
     Txrx_Tape *tapePtr;
     char *buf;
{
    if (tapePtr->inputAvailable) {
        tapePtr->inputAvailable = 0;
        buf[0] = (char)tapePtr->inputChar;
        return 1;
    }
    buf[0] = tapePtr->inText[tapePtr->inPos ++];
    return buf[0] != '\0';
}

/*
 *----------------------------------------------------------------------
 *
 * UngetChar --
 *
 *    Put back in input stream a character.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Character is stored locally and will be available on
 *      the next call to GetChar.
 *
 *----------------------------------------------------------------------
 */

static void
UngetChar(tapePtr, ch)
     Txrx_Tape *tapePtr;
     int ch;
{
    tapePtr->inputChar = (char)ch;
    tapePtr->inputAvailable = 1;
}

/*
 *----------------------------------------------------------------------
 *
 * MatchKeyword --
 *
 *    Checks if an identifier is a reserved RPC/XDR language keyword.
 *
 * Results:
 *    1 if it is, 0 otherwise.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
MatchKeyword(id, len)
     char *id;
     int len;
{
    
    if (len > 9 || len < 3)
            return 0;
    
    switch (len) {
      case 3:
        if (strncmp(id, "int", 3) == 0)
                return 1;
        return 0;
      case 4:
        if (strncmp(id, "bool", 4) == 0
            || strncmp(id, "case", 4) == 0
            || strncmp(id, "enum", 4) == 0
            || strncmp(id, "long", 4) == 0
            || strncmp(id, "void", 4) == 0)
                return 1;
        return 0;
      case 5:
        if (strncmp(id, "const", 5) == 0
            || strncmp(id, "float", 5) == 0
            || strncmp(id, "hyper", 5) == 0
            || strncmp(id, "short", 5) == 0
            || strncmp(id, "union", 5) == 0)
                return 1;
        return 0;
      case 6:
        if (strncmp(id, "double", 6) == 0
            || strncmp(id, "opaque", 6) == 0
            || strncmp(id, "string", 6) == 0
            || strncmp(id, "struct", 6) == 0
            || strncmp(id, "switch", 6) == 0)
                return 1;
        return 0;
      case 7:
        if (strncmp(id, "default", 7) == 0
            || strncmp(id, "typedef", 7) == 0
            || strncmp(id, "program", 7) == 0
            || strncmp(id, "version", 7) == 0)
                return 1;
        return 0;
      case 8:
        if (strncmp(id, "unsigned", 8) == 0)
                return 1;
        return 0;
      case 9:
        if (strncmp(id, "quadruple", 9) == 0)
                return 1;
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Parse --
 *
 *    Parse a protocol description file and build a tape out of it.
 *
 * Results:
 *    1 if succesfull, 0 if a lexical error occured. In case of
 *      error the result of the interpreter contains more information.
 *
 * Side effects:
 *    Tape is allocated and built.
 *
 *----------------------------------------------------------------------
 */

static int
Parse(tapePtr)
     Txrx_Tape *tapePtr;
{
    /* Initialize data structures - symbol tables maintained on tape. */
    
    /* Reset lexical scanner, otherwise state from previous parses
     * may be used in current parses.
     */
    tapePtr->inputAvailable = 0;
    tapePtr->inputChar = 0;
    tapePtr->lexemeAvailable = 0;
    /* Start parsing. */
    if (Specification(tapePtr) == TCL_ERROR) {
        AddLineErrorInfo(tapePtr);
        return TCL_ERROR;
    }
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * AddLineErrorInfo --
 *
 *    In case of parse error, add information about the line and
 *    column where the error occured.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Add information to interp->result.
 *
 *----------------------------------------------------------------------
 */

static void
AddLineErrorInfo(tapePtr)
     Txrx_Tape *tapePtr;
{
    int col, row;
    int i;
    char buf[80];
    /* 
     * Scan from the current position in the buffer and 
     * count the characters to the first end of line and
     * the number of end of line characters to the
     * beginning of the buffer.
     */
    for (i = tapePtr->inPos, col = 0; 
         i >= 0 && tapePtr->inText[i] != '\n'; i --, col ++);
    /*
     * Column 0 doesn't make much sense.
     */
    if (col > 1) col --;
    if (col == 0) col ++;            
    for (i = tapePtr->inPos, row = 1; i >= 0; i --) {
        row += (tapePtr->inText[i] == '\n');
    }
    sprintf(buf, "error at line %d, column %d\n", row, col);
    Tcl_AppendResult(tapePtr->interp, buf, (char *)NULL);
}

/*
 *----------------------------------------------------------------------
 *
 * Specification --
 *
 *    Parse a protocol specification.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Translates a protocol definition to tape format.
 *
 *----------------------------------------------------------------------
 */

static int
Specification(tapePtr)
     Txrx_Tape *tapePtr;
{
    Token token;
    int res = TCL_OK;
    
    while (res == TCL_OK && Scan(tapePtr, &token, UNKNOWN) == TCL_OK) {
        if (token.type == EOFFOUND) { /* Finished processing file. */
            return TCL_OK;
        }
        if (token.type != KEYWRD) {
            goto error;
        }

        if (strcmp(token.identifier, "const") == 0) {
            /* Parse a constant definition. */
            res = ConstantDef(tapePtr);
            continue;
        }
        if (   strcmp(token.identifier, "typedef") == 0 
            || strcmp(token.identifier, "enum") == 0
            || strcmp(token.identifier, "struct") == 0
            || strcmp(token.identifier, "union") == 0) {
            /* Parse a type definition. */
            res = TypeDef(tapePtr, &token);
            continue;
        }
        if ( strcmp(token.identifier, "program") == 0) {
            /* Parse a program definition. */
            res = ProgramDef(tapePtr);
            continue;
        }
      error:
        Tcl_AppendResult(tapePtr->interp, "expected one of ",
                         "\"typedef\", \"enum\", \"struct\", \"union\", ",
                         "\"const\" or \"program\"\n",
                         (char *)NULL);
        res = TCL_ERROR;
    }

    return TCL_ERROR;
}



/*
 *----------------------------------------------------------------------
 *
 * ConstantDef --
 *
 *    Process a constant definition.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Constant is added to symbol table.
 *
 *----------------------------------------------------------------------
 */

static int
ConstantDef(tapePtr)
     Txrx_Tape *tapePtr;
{
    Token idToken;
    Token specialToken;
    Token valueToken;
    int isNewEntry = 0;
    SymbolData *constSymbol;
    Tcl_HashEntry *entryPtr;
    long constValue;

    if (   Scan(tapePtr, &idToken, IDENT) == TCL_ERROR 
        || Scan(tapePtr, &specialToken, EQUAL) == TCL_ERROR
        || Scan(tapePtr, &valueToken, UNKNOWN) == TCL_ERROR
        || GetValue(tapePtr, &valueToken, &constValue) == TCL_ERROR
        || Scan(tapePtr, &specialToken, SEMICOLUMN) == TCL_ERROR) {
        return TCL_ERROR;
    }
    
    /* 
     * Constant definition is syntactically correct, add entry to hash
     * table.
     */
    entryPtr = Tcl_CreateHashEntry(&tapePtr->constTypeTable, 
                                   idToken.identifier, &isNewEntry);
    if (!isNewEntry) {
        Tcl_AppendResult(tapePtr->interp, "duplicate constant name \"",
                         idToken.identifier, "\"\n", (char *)NULL);
        return TCL_ERROR;
    }    
    
    constSymbol = InitSymbol();
    constSymbol->constValue = constValue;
    constSymbol->type = SYM_CONST;
    
    Tcl_SetHashValue(entryPtr, (ClientData) constSymbol);
    return TCL_OK;
} 

/*
 *----------------------------------------------------------------------
 *
 * TypeDef --
 *
 *    Process a type definition.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Type is added to symbol table.
 *
 *----------------------------------------------------------------------
 */

static int
TypeDef(tapePtr, token)
     Txrx_Tape *tapePtr;
     Token *token;
{
    Token idToken, specialToken;
    char typeName[MAXIDENTLEN + 1];
    SymbolData *symbolPtr;
    int res, isNewEntry = 0;
    Tcl_HashEntry *entryPtr;

    /* Initialize type name to a zero length string - if the 
     * type name is not modified then the declaration defines
     * no type, for example "typedef void".
     */
    typeName[0] = '\0';

    /*
     * Fill in symbol with blank data. 
     */
    symbolPtr = InitSymbol();    
    
    if (strcmp(token->identifier, "typedef") == 0) {
        if (Declaration(tapePtr, symbolPtr, typeName) == TCL_ERROR) {
            goto error;
        }
        if (typeName[0] == '\0') { /* No type was defined. */
            FreeSymbol(symbolPtr);
            return Scan(tapePtr, &specialToken, SEMICOLUMN);
        }
        entryPtr = Tcl_CreateHashEntry(&tapePtr->constTypeTable, 
                                       typeName, &isNewEntry);
        if (!isNewEntry) { /* Type redefinition. */
            Tcl_AppendResult(tapePtr->interp, "duplicate type name \"",
                             typeName, "\"\n", (char *)NULL);
            goto error;
        }    
    } else {
        /* The token is one of "enum", "struct" or "union". */
        if (Scan(tapePtr,&idToken, IDENT) == TCL_ERROR) {
            goto error;
        }
        /* Create entry in symbol table for recursive use of the name. */
        strcpy(typeName, idToken.identifier);
        entryPtr = Tcl_CreateHashEntry(&tapePtr->constTypeTable, 
                                       typeName, &isNewEntry);
        if (!isNewEntry) {
            Tcl_AppendResult(tapePtr->interp, "duplicate type name \"",
                             typeName, "\"\n", (char *)NULL);
            goto error;
        }    
        if (strcmp(token->identifier, "enum") == 0) {
            res = EnumBody(tapePtr, symbolPtr);
        } else if (strcmp(token->identifier, "struct") == 0) {
            res = StructBody(tapePtr, symbolPtr);
        } else if (strcmp(token->identifier, "union") == 0) {
            res = UnionBody(tapePtr, symbolPtr);
        } else {
            res = TCL_ERROR;
        }
        
        if (res == TCL_ERROR) {
            /* Delete entry from symbol table. */
            goto error1;
        }
    }

    if (Scan(tapePtr,&specialToken, SEMICOLUMN) == TCL_ERROR) {
        goto error1;
    }
    /* 
     * Parse was succesful. Add symbol symbol data to
     * the entry in the type table.
     */
    Tcl_SetHashValue(entryPtr, (ClientData) symbolPtr);
    
    return TCL_OK;

  error1:
    Tcl_DeleteHashEntry(entryPtr);
  error:
    FreeSymbol(symbolPtr);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * ProgramDef --
 *
 *    Parse a program definition.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Program, versions, procedures are added to symbol table.
 *
 *----------------------------------------------------------------------
 */

static int
ProgramDef(tapePtr)
     Txrx_Tape *tapePtr;
{
    Token idToken, specialToken, valueToken;
    SymbolData *progSymbolPtr;
    Tcl_HashTable *progTablePtr;
    Tcl_HashSearch searchInfo;
    Tcl_HashEntry *entryPtr;
    char versionName[MAXIDENTLEN + 1];
    SymbolData *versionSymbolPtr;
    long constValue;

    if (Scan(tapePtr,&idToken, IDENT) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected program name\n", (char *)NULL);
        return TCL_ERROR;
    }
    if (Scan(tapePtr,&specialToken, LCURLY) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected \"{\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    progTablePtr = (Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(progTablePtr, TCL_STRING_KEYS);
    do {
        versionSymbolPtr = InitSymbol();
        versionName[0] = '\0';
        /*
         * Parse each version and add it to the version table.
         */
        if (VersionDef(tapePtr, versionSymbolPtr, versionName) == TCL_ERROR) {
            goto error1;
        }

        if (NumberInTable(versionSymbolPtr->constValue,
                          progTablePtr, SYM_VERSION)) {
            Tcl_AppendResult(tapePtr->interp, "version number already used\n", 
                             (char *) NULL);
            goto error1;
        }
        if (AddToTable(tapePtr, progTablePtr, versionName, 
                       (ClientData)versionSymbolPtr) == TCL_ERROR) {
            goto error1;
        }
        if (Scan(tapePtr,&specialToken, RCURLY) == TCL_ERROR) {
            Tcl_ResetResult(tapePtr->interp);
            UnScan(tapePtr,&specialToken);
            continue;
        }
        /* Scanned the closing brace. */
        if (   Scan(tapePtr,&specialToken, EQUAL) == TCL_ERROR 
            || Scan(tapePtr,&valueToken, UNKNOWN) == TCL_ERROR 
            || GetValue(tapePtr, &valueToken, &constValue) == TCL_ERROR
            || Scan(tapePtr,&specialToken, SEMICOLUMN) == TCL_ERROR) {
            goto error;
        }
        if (NumberInTable((long)valueToken.value, &tapePtr->constTypeTable,
                     SYM_PROGRAM)) {
            Tcl_AppendResult(tapePtr->interp,
                             "duplicate program number\n", (char *)NULL);
            goto error;
        }
        progSymbolPtr = InitSymbol();
        /*
         * Add program to type table.
         */
        if (AddToTable(tapePtr,
                       &tapePtr->constTypeTable, idToken.identifier, 
                       (ClientData) progSymbolPtr) == TCL_ERROR) {
            FreeSymbol(progSymbolPtr);
            goto error;
        }

        progSymbolPtr->constValue = constValue;
        progSymbolPtr->hashTablePtr = progTablePtr;
        progSymbolPtr->type = SYM_PROGRAM;
        entryPtr = Tcl_FirstHashEntry(progTablePtr, &searchInfo);
        while (entryPtr) {
            /*
             * Store a reference to the program in the version.
             */
            versionSymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
            versionSymbolPtr->referenceValue = progSymbolPtr;
            entryPtr = Tcl_NextHashEntry(&searchInfo);
        }
        return TCL_OK;
    } while (1);
  error1:
    FreeSymbol(versionSymbolPtr);
  error:
    FreeSymbolTable(progTablePtr);
    ckfree((char *)progTablePtr);

    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * VersionDef --
 *
 *    Parse a version definition.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Version and component procedures are added to symbol table.
 *
 *----------------------------------------------------------------------
 */

static int
VersionDef(tapePtr, symbolPtr, name)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     char *name;
{
    Token specialToken, idToken, valueToken;
    SymbolData *procSymbolPtr;
    Tcl_HashTable *procTablePtr;
    char procName[MAXIDENTLEN + 1];
    long constValue;

    if (   Scan(tapePtr,&specialToken, KEYWRD) == TCL_ERROR
        || strcmp(specialToken.identifier, "version") != 0) {
        Tcl_AppendResult(tapePtr->interp, "expected \"version\" keyword\n", 
                         (char *)NULL);
        return TCL_ERROR;
    }
    if (Scan(tapePtr,&idToken, IDENT) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, 
                         "expected version name\n", (char *)NULL);
        return TCL_ERROR;
    }
    if (Scan(tapePtr,&specialToken, LCURLY) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected \"{\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    
    procTablePtr = (Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(procTablePtr, TCL_STRING_KEYS);
    do {
        procSymbolPtr = InitSymbol();
        procName[0] = '\0';
        if (ProcedureDef(tapePtr, procSymbolPtr, procName) == TCL_ERROR) {
            FreeSymbol(procSymbolPtr);
            goto error;
        }
        /*
         * Store a reference to the version in the procedure.
         */
        procSymbolPtr->referenceValue = symbolPtr;
        if (NumberInTable(procSymbolPtr->constValue,
                          procTablePtr, SYM_PROCEDURE)) {
            Tcl_AppendResult(tapePtr->interp, "procedure number already used\n", 
                             (char *)NULL);
            FreeSymbol(procSymbolPtr);
            goto error;
        }
        if (AddToTable(tapePtr, procTablePtr, procName, 
                       (ClientData)procSymbolPtr) == TCL_ERROR) {
            FreeSymbol(procSymbolPtr);
            goto error;
        }

        if (Scan(tapePtr, &specialToken, RCURLY) == TCL_ERROR) {
            Tcl_ResetResult(tapePtr->interp);
            UnScan(tapePtr,&specialToken);
            continue;
        }
        /* Scanned the closing brace. */
        if (   Scan(tapePtr, &specialToken, EQUAL) == TCL_ERROR 
            || Scan(tapePtr, &valueToken, UNKNOWN) == TCL_ERROR 
            || GetValue(tapePtr, &valueToken, &constValue) == TCL_ERROR
            || Scan(tapePtr, &specialToken, SEMICOLUMN) == TCL_ERROR) {
            goto error;
        }

        symbolPtr->constValue = constValue;
        symbolPtr->hashTablePtr = procTablePtr;
        symbolPtr->type = SYM_VERSION;
        strcpy(name, idToken.identifier);
        return TCL_OK;

    } while (1);

  error:
    FreeSymbolTable(procTablePtr);
    ckfree((char *)procTablePtr);

    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * ProcedureDef --
 *
 *    Parse a procedure definition.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Component arguments and type references are added to the symbol table.
 *
 *----------------------------------------------------------------------
 */

static int
ProcedureDef(tapePtr, symbolPtr, name)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     char *name;
{
    Token specialToken, idToken, valueToken;
    Tcl_HashTable *argTablePtr 
            = (Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    SymbolData *argSymbolPtr;
    char argName[MAXIDENTLEN + 1];
    SymbolData *resultSymbolPtr = InitSymbol();
    SymbolData *constSymbolPtr;
    int argc = 0;
    char argcName[10];
    long constValue;

    Tcl_InitHashTable(argTablePtr, TCL_STRING_KEYS);
    
    if (   Scan(tapePtr, &specialToken, UNKNOWN) == TCL_ERROR 
        || TypeSpecifier(tapePtr, 
                         resultSymbolPtr, &specialToken) == TCL_ERROR) {
        goto error;
    }
    AddToTable(tapePtr, argTablePtr, "0", (ClientData)resultSymbolPtr);
    
    if (Scan(tapePtr, &idToken, IDENT) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, 
                         "expected procedure name\n", (char *)NULL);
        goto error;
    }
    if (Scan(tapePtr,&specialToken, LPAREN) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected \"(\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    
    do {
        argc ++; sprintf(argcName, "%d", argc);
        argSymbolPtr = InitSymbol();
        argName[0] = '\0';
        if (   Scan(tapePtr,&specialToken, UNKNOWN) == TCL_ERROR
            || TypeSpecifier(tapePtr,
                             argSymbolPtr, &specialToken) == TCL_ERROR) {
            FreeSymbol(argSymbolPtr);
            goto error;
        }
        if (AddToTable(tapePtr,
                       argTablePtr, 
                       argcName, 
                       (ClientData)argSymbolPtr) == TCL_ERROR) {
            FreeSymbol(argSymbolPtr);
            Tcl_AppendResult(tapePtr->interp, 
                             "internal error argument ", argcName,
                             ", ", argName, (char *)NULL);
            goto error;
        }
        if (Scan(tapePtr,&specialToken, UNKNOWN) == TCL_ERROR) {
            goto error;
        }
        if (specialToken.type == COMMA) {
            continue;
        }
        if (specialToken.type != RPAREN) {
            Tcl_AppendResult(tapePtr->interp, "expected \",\" or \")\"\n", 
                             (char *)NULL);
            goto error;
        }
        /* Done with the arguments. */
        if (   Scan(tapePtr,&specialToken, EQUAL) == TCL_ERROR
            || Scan(tapePtr,&valueToken, UNKNOWN) == TCL_ERROR
            || GetValue(tapePtr, &valueToken, &constValue) == TCL_ERROR
            || Scan(tapePtr,&specialToken, SEMICOLUMN) == TCL_ERROR) {
            goto error;
        }

        /* 
         * (Try to) add the procedure name as a constant to
         * the global type table. This is a non standard 
         * extension, thus if a constant with the name of the
         * procedure has already been defined, there is no
         * error.
         *
         * Keep a pointer to the program symbol in the constant
         * symbol so that retrieval is easy.
         */
        constSymbolPtr = InitSymbol();
        constSymbolPtr->constValue = constValue;
        constSymbolPtr->referenceValue = symbolPtr;
        constSymbolPtr->type = SYM_CONST;
        if (AddToTable(tapePtr, &tapePtr->constTypeTable, idToken.identifier, 
                       (ClientData) constSymbolPtr) == TCL_ERROR) {
            FreeSymbol(constSymbolPtr);
            Tcl_ResetResult(tapePtr->interp);
        }

        /* Input was syntactically correct, prepare procedure symbol. */
        symbolPtr->constValue = constValue;
        symbolPtr->hashTablePtr = argTablePtr;
        symbolPtr->type = SYM_PROCEDURE;
        symbolPtr->identValue = (char *)argc;
        strcpy(name, idToken.identifier);
        return TCL_OK;

    } while (1);

  error:
    FreeSymbolTable(argTablePtr);
    ckfree((char *)argTablePtr);

    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * NumberInTable --
 *
 *    Search through a symbol table and for each symbol  
 *    matching the type check if the number given as parameter 
 *    is not the number stored in the constValue field (or 
 *    as hash value directly for CONSTANTS.
 *
 * Results:
 *    1 if the number was IS found, 0 if it was not found in the table.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int 
NumberInTable(n, tablePtr, symType)
     long n;
     Tcl_HashTable *tablePtr;
     int symType;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    SymbolData *symbolPtr;

    entryPtr = Tcl_FirstHashEntry(tablePtr, &searchInfo);
    while (entryPtr) {
        symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (symType == symbolPtr->type) {
            if (symbolPtr->constValue == n)  {
                return 1;
            }
        }
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Declaration --
 *
 *    Parse a declaration specification.
 *
 * Results:
 *    A standard TCL result. The root symbol for the parse tree of 
 *    is an output parameter. The name of the type will be stored 
 *    in the array of characters passed as parameter.
 *
 * Side effects:
 *    As required.
 *
 *----------------------------------------------------------------------
 */

static int
Declaration(tapePtr, symbolPtr, typeName)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     char *typeName;
{
    SymbolData *pointerSymbolPtr;
    Token token;

    if (Scan(tapePtr,&token, UNKNOWN) == TCL_ERROR) {
        return TCL_ERROR;
    }

    if (   token.type == KEYWRD
        && strcmp(token.identifier, "void") == 0) {
        /* Void symbol, already initialized. */
        symbolPtr->codeAddress = TVOID;
        return TCL_OK;
    }
    if (   token.type == KEYWRD 
        && strcmp(token.identifier, "string") == 0) {
        return StringDeclaration(tapePtr, symbolPtr, typeName);
    } else if (   token.type == KEYWRD
               && strcmp(token.identifier, "opaque") == 0) {
        return OpaqueDeclaration(tapePtr, symbolPtr, typeName);
    } else {
        if (TypeSpecifier(tapePtr, symbolPtr, &token) == TCL_ERROR) {
            return TCL_ERROR;
        }
        /* 
         * Need to separate between: pointers(*) and new type name.
         */
        if (Scan(tapePtr,&token, UNKNOWN) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (token.type == STAR) {
            if (Scan(tapePtr,&token, IDENT) == TCL_ERROR) {
                return TCL_ERROR;
            }
            /*
             * Make the symbol a reference symbol.
             */
            pointerSymbolPtr = InitSymbol();
            *pointerSymbolPtr = *symbolPtr;
            symbolPtr->codeAddress = 0;
            symbolPtr->constValue = 0l;
            symbolPtr->identValue = NULL;
            symbolPtr->referenceValue = pointerSymbolPtr;
            symbolPtr->defaultValue = NULL;
            symbolPtr->hashTablePtr = NULL;
            symbolPtr->type = SYM_POINTER;
            /* Store the new type name. */
            strcpy(typeName, token.identifier);
            return TCL_OK;
        } else if (token.type == IDENT) {
            return NamedType(tapePtr, symbolPtr, &token, typeName);
        } 
        Tcl_AppendResult(tapePtr->interp, "expected identifier or \"*\"\n", 
                         (char *)NULL);
    }
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * TypeSpecifier --
 *
 *    Parse a type specifier.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    The symbol is initialized with the type parsed.
 *
 *----------------------------------------------------------------------
 */

static int
TypeSpecifier(tapePtr, symbolPtr, token)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     Token *token;
{
    if (token->type == KEYWRD) {
        if (strcmp(token->identifier, "void") == 0) {
            /* Void symbol, already initialized. */
            symbolPtr->codeAddress = TVOID;
            return TCL_OK;
        }
        if (strcmp(token->identifier, "unsigned") == 0) {
            /* Get on to the next lexeme. */
            if (Scan(tapePtr,token, UNKNOWN) == TCL_ERROR) {
                return TCL_ERROR;
            } 
            if (token->type == STAR || token->type == IDENT) {
                symbolPtr->type = SYM_UNSIGNED;
                symbolPtr->codeAddress = TUINT;
                UnScan(tapePtr,token);
                return TCL_OK;
            }
            if (   token->type != KEYWRD 
                || (   strcmp(token->identifier, "int") != 0
                    && strcmp(token->identifier, "hyper") != 0)) {
                Tcl_AppendResult(tapePtr->interp, "\"unsigned\" must be followed by ",
                                 "\"int\", \"hyper\", \"*\" or identifier\n",
                                 (char *)NULL);
                return TCL_ERROR;
            }
            if (strcmp(token->identifier, "int") == 0) {
                symbolPtr->type = SYM_UNSIGNED;
                symbolPtr->codeAddress = TUINT;
            } else if (strcmp(token->identifier, "hyper") == 0) {
                symbolPtr->type = SYM_UNSIGNEDHYPER;
                symbolPtr->codeAddress = TUHYPER;
            }
            return TCL_OK;
        }
        if (   strcmp(token->identifier, "int") == 0
            || strcmp(token->identifier, "long") == 0) {
            symbolPtr->type = SYM_INT;
            symbolPtr->codeAddress = TINT;
        } else if (strcmp(token->identifier, "hyper") == 0) {
            symbolPtr->type = SYM_HYPER;
            symbolPtr->codeAddress = THYPER;
        } else if (strcmp(token->identifier, "bool") == 0) {
            symbolPtr->type = SYM_BOOL;
            symbolPtr->codeAddress = TBOOL;
        } else if (strcmp(token->identifier, "float") == 0) {
            symbolPtr->type = SYM_FLOAT;
            symbolPtr->codeAddress = TFLOAT;
        } else if (strcmp(token->identifier, "double") == 0) {
            symbolPtr->type = SYM_DOUBLE;
            symbolPtr->codeAddress = TDOUBLE;
        } else if (strcmp(token->identifier, "quadruple") == 0) {
            symbolPtr->type = SYM_QUADRUPLE;
            symbolPtr->codeAddress = TQUAD;
        } else if (strcmp(token->identifier, "enum") == 0) {
            if (EnumBody(tapePtr, symbolPtr) == TCL_ERROR) {
                return TCL_ERROR;
            }
        } else if (strcmp(token->identifier, "struct") == 0) {
            if (StructBody(tapePtr, symbolPtr) == TCL_ERROR) {
                return TCL_ERROR;
            }
        } else if (strcmp(token->identifier, "union") == 0) {
            if (UnionBody(tapePtr, symbolPtr) == TCL_ERROR) {
                return TCL_ERROR;
            }
        } else if (strcmp(token->identifier, "string") == 0) {
            symbolPtr->type = SYM_STRING;
            symbolPtr->identValue = (char *)1;
            symbolPtr->constValue = 0;
        } else if (strcmp(token->identifier, "opaque") == 0) {
            symbolPtr->type = SYM_OPAQUE;
            symbolPtr->identValue = (char *)1;
            symbolPtr->constValue = 0;
        } else {
            Tcl_AppendResult(tapePtr->interp, "invalid keyword \"",
                             token->identifier, "\"", (char *)NULL);
            return TCL_ERROR;
        }
    } else if (token->type == IDENT) {
        symbolPtr->type = SYM_TYPE;
        symbolPtr->identValue = (char *)ckalloc(token->length + 1);
        strcpy(symbolPtr->identValue, token->identifier);
    }
        
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * StringDeclaration --
 *
 *    Parse a string declaration.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    Symbol data is changed to store string information.
 *
 *----------------------------------------------------------------------
 */

static int
StringDeclaration(tapePtr, symbolPtr, typeName)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     char *typeName;
{
    Token idToken;
    Token specialToken;

    int isVar;
    long value;
    
    if (Scan(tapePtr,&idToken, IDENT) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected identifier\n", (char *)NULL);
        return TCL_ERROR;
    }
    if (Scan(tapePtr,&specialToken, LANGLE) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected \"<\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    if (GetArraySize(tapePtr, &specialToken, &value, &isVar) == TCL_ERROR) {
        return TCL_ERROR;
    }
    /*
     * Fill in the symbol with information.
     */
    symbolPtr->constValue = value;
    symbolPtr->identValue = (char *)isVar;
    symbolPtr->type = SYM_STRING;
    strcpy(typeName, idToken.identifier);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * OpaqueDeclaration --
 *
 *    Parse an opaque data declaration.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    Symbol data is changed to store information about the data definition.
 *
 *----------------------------------------------------------------------
 */

static int
OpaqueDeclaration(tapePtr, symbolPtr, typeName)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     char *typeName;
{
    Token idToken;
    Token specialToken;

    int isVar;
    long value;
    
    if (Scan(tapePtr,&idToken, IDENT) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, 
                         "expected identifier\n", (char *)NULL);
        return TCL_ERROR;
    }
    if (Scan(tapePtr,&specialToken, UNKNOWN) == TCL_ERROR ||
        (specialToken.type != LSQUARE && specialToken.type != LANGLE)) {
        Tcl_AppendResult(tapePtr->interp, 
                         "expected \"<\" or \"[\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    if (GetArraySize(tapePtr, &specialToken, &value, &isVar) == TCL_ERROR) {
        return TCL_ERROR;
    }
    /*
     * Fill in the symbol with information.
     */
    symbolPtr->constValue = value;
    symbolPtr->identValue = (char *)isVar;
    symbolPtr->type = SYM_OPAQUE;
    strcpy(typeName, idToken.identifier);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * NamedType --
 *
 *    Parse a named type like an array, a variable length array or a
 *    plain type.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    Symbol data is updated with information about the named type.
 *
 *----------------------------------------------------------------------
 */

static int
NamedType(tapePtr, symbolPtr, token, typeName)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     Token *token;
     char *typeName;
{
    Token specialToken;
    long arraySize;
    int isVar;
    SymbolData *arrayPtr;

    /* 
     * Parameter token contains an identifier represnting the name of the type.
     * Parameter symbolPtr contains a description of a type.
     */
    if (Scan(tapePtr,&specialToken, UNKNOWN) == TCL_ERROR) {
        return TCL_ERROR;
    }
    if (specialToken.type != LSQUARE && specialToken.type != LANGLE) {
        /* Not an array, put the lexeme back. */
        UnScan(tapePtr, &specialToken);
    } else {
        if (GetArraySize(tapePtr, 
                         &specialToken, &arraySize, &isVar) == TCL_ERROR) {
            return TCL_ERROR;
        }
        /* 
         * Allocate a symbol to describe the array. Copy element
         * of array description into it.
         */
        arrayPtr = InitSymbol();
        *arrayPtr = *symbolPtr;
        symbolPtr->constValue = arraySize;
        symbolPtr->identValue = (char *)isVar;
        symbolPtr->referenceValue = arrayPtr;
        symbolPtr->codeAddress = 0;
        symbolPtr->type = SYM_ARRAY;
        symbolPtr->defaultValue = NULL;
        symbolPtr->hashTablePtr = NULL;
    } 

    /* Save the type name. */
    strcpy(typeName, token->identifier);

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GetValue --
 *
 *    Get the numeric value for a token. Token is either a numeric
 *    constant or a constant already defined in the type table.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int 
GetValue(tapePtr, token, value)
     Txrx_Tape *tapePtr;
     Token *token;
     long *value;
{
    Tcl_HashEntry *entryPtr;
    SymbolData *symbolPtr;
 
    /* Check the token - it's either a number or a constant name. */
    if (token->type == VALUE) {
        *value = (long)token->value;
    } else if (token->type == IDENT) {
        /* Must be a numeric constant name - should be in
         * the constants and types symbol table.
         */
        entryPtr = Tcl_FindHashEntry(&tapePtr->constTypeTable,
                                     token->identifier);
        if (entryPtr == NULL) {
            Tcl_AppendResult(tapePtr->interp, "no such constant name \"",
                             token->identifier, "\"\n", (char *)NULL);
            return TCL_ERROR;
        }
        symbolPtr = (SymbolData *) Tcl_GetHashValue(entryPtr);
        if (symbolPtr->type != SYM_CONST) {
            Tcl_AppendResult(tapePtr->interp, "\"", token->identifier, 
                             "\" is not a constant\n", (char *)NULL);
            return TCL_ERROR;
        }
        *value = symbolPtr->constValue;
    } else {
        Tcl_AppendResult(tapePtr->interp, "expected identifier or value\n",
                         (char *)NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GetArraySize --
 *
 *  Extract from the input a numeric value representing the size of an 
 *  array. Token is either a numeric
 *  value or a constant symbol, already in the constants and
 *  types table.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
GetArraySize(tapePtr, token, value, isVar)
     Txrx_Tape *tapePtr;
     Token *token;
     long *value;
     int *isVar;
{
    Token valToken, nextToken;

    switch (token->type) {
      case LANGLE:
        *isVar = 1;
        break;
      case LSQUARE:
        *isVar = 0;
        break;
      default:
        Tcl_AppendResult(tapePtr->interp, 
                         "expected \"[\" or \"<\"\n", (char *)NULL);
        return TCL_ERROR;
    }

    if (Scan(tapePtr,&valToken, UNKNOWN) == TCL_ERROR) {
        return TCL_ERROR;
    }
    
    if (valToken.type == IDENT || valToken.type == VALUE) {
        if (GetValue(tapePtr, &valToken, value) == TCL_ERROR) {
            return TCL_ERROR;
        }
    } else {
        /*
         * Put the token back as it is not a value.
         */
        *value = 0;
        UnScan(tapePtr,&valToken);
    }

    if (Scan(tapePtr, &nextToken, UNKNOWN) == TCL_ERROR) {
        return TCL_ERROR;
    }
    if (nextToken.type != RANGLE && nextToken.type != RSQUARE) {
        Tcl_AppendResult(tapePtr->interp, "expected \"]\" or \">\"\n", (char *)NULL);
        return TCL_ERROR;
    }

    if (nextToken.type == RANGLE && *isVar != 1) {
        Tcl_AppendResult(tapePtr->interp, "expected \"]\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    if (nextToken.type == RSQUARE && *isVar != 0) {
        Tcl_AppendResult(tapePtr->interp, "expected \">\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * AddToTable --
 *
 *    Add a new entry to a hash table. Key is not in the table.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
AddToTable(tapePtr, tablePtr, id, clientData)
     Txrx_Tape *tapePtr;
     Tcl_HashTable *tablePtr;
     char *id;
     ClientData clientData;
{
    int isNew;
    Tcl_HashEntry *entryPtr;
    
    entryPtr = Tcl_CreateHashEntry(tablePtr, id, &isNew);
    if (!isNew) {
        Tcl_AppendResult(tapePtr->interp, 
                         "duplicate name \"", id,
                         "\" in declaration\n", (char *)NULL);
        return TCL_ERROR;
    }
    Tcl_SetHashValue(entryPtr, clientData);
    return TCL_OK;
} 

/*
 *----------------------------------------------------------------------
 *
 * EnumBody --
 *
 *    Process an enum definition.
 *
 * Results:
 *    A standard TCL result.
 *
 * Parameters:
 *      Symbol to process.
 *
 * Side effects:
 *    Symbol hash table is initialized with enumeration values.
 *
 *----------------------------------------------------------------------
 */

static int 
EnumBody(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Token special;
    Token enumId;
    Token enumValue;
    Tcl_HashTable *enumTablePtr;
    long value;
    long count;           /* Number of elements in the table. */

    if (Scan(tapePtr,&special, UNKNOWN) == TCL_ERROR) {
        return TCL_ERROR;
    }
    if (special.type == IDENT) {
        /* 
         * Non standard extension, assume it's a type reference.
         */
        symbolPtr->type = SYM_TYPE;
        symbolPtr->identValue = (char *)ckalloc(special.length + 1);
        strcpy(symbolPtr->identValue, special.identifier);
        return TCL_OK;
    }
    if (special.type != LCURLY) {
        Tcl_AppendResult(tapePtr->interp, "expected \"{\"\n", (char *)NULL);
        return TCL_ERROR;
    }

    enumTablePtr = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(enumTablePtr, TCL_STRING_KEYS);
    count = 0; 
    
    do {
        count ++;
        if (Scan(tapePtr,&enumId, IDENT) == TCL_ERROR
            || Scan(tapePtr,&special, EQUAL) == TCL_ERROR
            || Scan(tapePtr,&enumValue, UNKNOWN) == TCL_ERROR) {
            goto error;
        } 
        /*
         * Get the numeric value for the identifier.
         */
        if (GetValue(tapePtr, &enumValue, &value) == TCL_ERROR) {            
            goto error;
        }
        /*
         * Add pair to the hash table - identifier is key,
         * numeric value (which must evaluate to an int)
         * is the ClientData field of the entry.
         */
        if (AddToTable(tapePtr,
                       enumTablePtr, 
                       enumId.identifier, 
                       (ClientData)value) == TCL_ERROR) {
            goto error;
        }
        if (Scan(tapePtr,&special, UNKNOWN) == TCL_ERROR) {
            goto error;
        }
        if (special.type == COMMA) {
            continue;
        }
        if (special.type == RCURLY) {
            /* 
             * Fill the symbol received as parameter.
             */ 
            symbolPtr->hashTablePtr = enumTablePtr;
            symbolPtr->constValue = count;
            symbolPtr->type = SYM_ENUM;
            return TCL_OK;
        }
        Tcl_AppendResult(tapePtr->interp, "expected \",\" or \"}\"\n", (char *)NULL);
        goto error;
    } while (1);
  error:
    /*
     * Note: table values are integers.
     */
    Tcl_DeleteHashTable(enumTablePtr);
    ckfree((char *)enumTablePtr);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * StructBody --
 *
 *    Process a structurem definition.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    Structure fields and contents are parsed.
 *
 *----------------------------------------------------------------------
 */

static int
StructBody(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Token special;
    Tcl_HashTable *structTablePtr;
    Tcl_HashTable fieldTable;
    SymbolData *fieldSymbolPtr;
    char fieldName[MAXIDENTLEN + 1];
    int argc;
    char argcBuf[10];

    if (Scan(tapePtr, &special, UNKNOWN) == TCL_ERROR) {
        return TCL_ERROR;
    }
    if (special.type == IDENT) {
        /* 
         * Non standard extension, assume it's a type reference.
         */
        symbolPtr->type = SYM_TYPE;
        symbolPtr->identValue = (char *)ckalloc(special.length + 1);
        strcpy(symbolPtr->identValue, special.identifier);
        return TCL_OK;
    }
    
    if (special.type != LCURLY) {
        Tcl_AppendResult(tapePtr->interp, "expected \"{\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    structTablePtr = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(structTablePtr, TCL_STRING_KEYS);
    Tcl_InitHashTable(&fieldTable, TCL_STRING_KEYS);
    argc = 0;

    do {
        argc ++; sprintf(argcBuf, "%d", argc);
        /*
         * Create a symbol for the field. 
         */
        fieldSymbolPtr = InitSymbol();
        
        fieldName[0] = '\0';
        if (Declaration(tapePtr, fieldSymbolPtr, fieldName) == TCL_ERROR) {
            goto error1;
        }
        /*
         * If the fieldName is empty then there must have been an error.
         */
        if (fieldName[0] == '\0') {
            Tcl_AppendResult(tapePtr->interp, "field name in struct is empty\n",
                             (char *)NULL);
            goto error1;
        }
        /*
         * Add field name and its defintion to the hash table.
         * Check for duplicate field names.
         */
        if (AddToTable(tapePtr,
                       &fieldTable, 
                       fieldName, 
                       (ClientData)NULL) == TCL_ERROR) {
            goto error1;
        }
        
        if (AddToTable(tapePtr, 
                       structTablePtr, 
                       argcBuf, 
                       (ClientData)fieldSymbolPtr) == TCL_ERROR) {
            Tcl_AppendResult(tapePtr->interp, 
                             "internal error while processing ",
                             "procedure arguments\n", (char *)NULL);
            goto error1;
        }
        if (Scan(tapePtr,&special, SEMICOLUMN) == TCL_ERROR) {
            Tcl_AppendResult(tapePtr->interp, "expected \";\"\n", (char *)NULL);
            goto error;
        }
        if (Scan(tapePtr,&special, UNKNOWN) == TCL_ERROR) {
            goto error;
        }
        if (special.type != RCURLY) {
            UnScan(tapePtr,&special);
            continue;
        }
        /* 
         * Done with the body of the struct, setup symbol pointer.
         */
        symbolPtr->constValue = argc;
        symbolPtr->hashTablePtr = structTablePtr;
        symbolPtr->type = SYM_STRUCT;
        
        Tcl_DeleteHashTable(&fieldTable);

        return TCL_OK;

    } while (1);
  error1:
    FreeSymbol(fieldSymbolPtr);
  error:
    /*
     * Free al symbols in the symbol table.
     */
    FreeSymbolTable(structTablePtr);
    ckfree((char *)structTablePtr);
    Tcl_DeleteHashTable(&fieldTable);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * UnionBody --
 *
 *    Process an union definition.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    Fields, names definitions are added to symbol tables.
 *
 *----------------------------------------------------------------------
 */

static int
UnionBody(tapePtr, symbolPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
{
    Token specialToken, labelToken;
    Tcl_HashTable *unionTablePtr;
    Tcl_HashTable fieldTable;
    SymbolData *discrSymbolPtr = NULL;
    SymbolData *caseSymbolPtr = NULL;
    CaseData *caseDataPtr = NULL;
    CaseData *defaultCasePtr = NULL;
    char discrTypeName [MAXIDENTLEN + 1];
    char typeName [MAXIDENTLEN + 1];

    int isDefault;
    int hasDefault = 0;

    if (Scan(tapePtr,&specialToken, UNKNOWN) == TCL_ERROR) {
        return TCL_ERROR;
    }
    if (specialToken.type == IDENT) {
        /* 
         * Non standard extension, assume it's a type reference.
         */
        symbolPtr->type = SYM_TYPE;
        symbolPtr->identValue = (char *)ckalloc(specialToken.length + 1);
        strcpy(symbolPtr->identValue, specialToken.identifier);
        return TCL_OK;
    }
    
    if ( specialToken.type != KEYWRD
        || strcmp(specialToken.identifier, "switch") != 0) {
        Tcl_AppendResult(tapePtr->interp, "expected \"switch\" keyword\n", 
                         (char *)NULL);
        return TCL_ERROR;
    }
    
    unionTablePtr = (Tcl_HashTable *)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(unionTablePtr, TCL_STRING_KEYS);
    Tcl_InitHashTable(&fieldTable, TCL_STRING_KEYS);
    
    discrSymbolPtr = InitSymbol();
    discrTypeName[0] = '\0';
    if (Discriminant(tapePtr, discrSymbolPtr, discrTypeName) == TCL_ERROR) {
        goto error;
    }
    if (Scan(tapePtr,&specialToken, LCURLY) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected \"{\"\n", (char *)NULL);
        goto error;
    }
    
    do {
        caseSymbolPtr = InitSymbol(); typeName[0] = '\0'; isDefault = 0;
        if (Case(tapePtr, 
                 caseSymbolPtr,typeName,&isDefault,&labelToken) == TCL_ERROR) {
            FreeSymbol(caseSymbolPtr); goto error;
        }
        if (typeName[0] != '\0' && AddToTable(tapePtr, 
                                              &fieldTable, 
                                              typeName, 
                                              (ClientData)NULL) == TCL_ERROR) {
            FreeSymbol(caseSymbolPtr); 
            goto error;
        }
        
        
        caseDataPtr = InitCase(caseSymbolPtr, &labelToken, typeName);
        if (isDefault) {
            if (hasDefault) {
                Tcl_AppendResult(tapePtr->interp, 
                                 "multiple \"default\" labels\n",
                                 (char *)NULL);
                goto error1;
            }
            hasDefault = 1;
          defaultCasePtr = caseDataPtr;
        } else {
            if (AddToTable(tapePtr, unionTablePtr, labelToken.identifier, 
                           (ClientData) caseDataPtr) == TCL_ERROR) {
                goto error1;
            }
        }
        /* Scan the semicolumn. */
        if (Scan(tapePtr,&specialToken, SEMICOLUMN) == TCL_ERROR) {
            Tcl_AppendResult(tapePtr->interp, "expected \";\"\n", (char *)NULL);
            goto error;
        }
        if (Scan(tapePtr,&specialToken, UNKNOWN) == TCL_ERROR) {
            goto error;
        }
        if (specialToken.type == RCURLY) {
            /*
             * Done. Fill in the current symbol.
             */
            symbolPtr->defaultValue = hasDefault?(char *)defaultCasePtr:NULL;
            symbolPtr->identValue = (char *)ckalloc(strlen(discrTypeName) + 1);
            strcpy(symbolPtr->identValue, discrTypeName);
            symbolPtr->referenceValue = discrSymbolPtr;
            symbolPtr->constValue = (long)hasDefault;
            symbolPtr->hashTablePtr = unionTablePtr;
            symbolPtr->type = SYM_UNION;

            Tcl_DeleteHashTable(&fieldTable);
            
            return TCL_OK;
        }
        /* 
         * Not a closing brace, put it back.
         */
        UnScan(tapePtr,&specialToken);
    } while (1);
    

  error1:
    FreeCase(caseDataPtr);
  error:
    if (hasDefault) {
        FreeCase(defaultCasePtr);
    }
    FreeCaseTable(unionTablePtr);
    ckfree((char *)unionTablePtr);
    if (discrSymbolPtr != NULL) {
        FreeSymbol(discrSymbolPtr);
    }
    Tcl_DeleteHashTable(&fieldTable);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * Case --
 *
 *    Parse a case label and declaration.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    Case label and declaration are stored.
 *
 *----------------------------------------------------------------------
 */

static int
Case(tapePtr, symbolPtr, namePtr, isDefault, labelPtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     char *namePtr;
     int *isDefault;
     Token *labelPtr;
{
    Token specialToken;
    
    if (Scan(tapePtr,&specialToken, KEYWRD) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, 
                         "expected \"case\" or \"default\" keyword\n",
                         (char *)NULL);
        return TCL_ERROR;
    }
    if (strcmp(specialToken.identifier, "case") == 0) {
        *isDefault = 0;
        if (Scan(tapePtr,labelPtr, UNKNOWN) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (labelPtr->type != IDENT && labelPtr->type != VALUE) {
            Tcl_AppendResult(tapePtr->interp, "case label must be a numeric constant",
                             " or an identifier\n", (char *)NULL);
            return TCL_ERROR;
        }
        if (labelPtr->type == VALUE) {
            /*
             * Prepare a string key in the identifier. 
             */
            sprintf(labelPtr->identifier, "%ld", (long) labelPtr->value);
        }
    } else if (strcmp(specialToken.identifier, "default") == 0) {
        *isDefault = 1;
    } else {
        Tcl_AppendResult(tapePtr->interp, 
                         "expected \"case\" or \"default\" keyword\n",
                         (char *)NULL);
        return TCL_ERROR;
    }
    
    if (Scan(tapePtr,&specialToken, COLUMN) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected \":\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    
    if (Declaration(tapePtr, symbolPtr, namePtr) == TCL_ERROR) {
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Discriminant --
 *
 *    Parse the discriminant field of a union.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Symbol is filled with information.
 *
 *----------------------------------------------------------------------
 */

static int
Discriminant(tapePtr, symbolPtr, namePtr)
     Txrx_Tape *tapePtr;
     SymbolData *symbolPtr;
     char *namePtr;
{
    Token token;
    
    if (Scan(tapePtr,&token, LPAREN) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected \"(\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    if (Declaration(tapePtr, symbolPtr, namePtr) == TCL_ERROR) {
        return TCL_ERROR;
    }
    /* 
     * The symbol and the name for the discriminant have been parsed
     * and stored locally.
     */
    if (Scan(tapePtr,&token, RPAREN) == TCL_ERROR) {
        Tcl_AppendResult(tapePtr->interp, "expected \")\"\n", (char *)NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FreeSymbol --
 *
 *    Free all data structures associated with a symbol definition.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Memory allocated by symbol is freed.
 *
 *----------------------------------------------------------------------
 */

static void
FreeSymbol(symbolPtr)
     SymbolData *symbolPtr;
{
    if (symbolPtr == NULL) {
        /*printf("internal error: freeing a NULL symbol.\n");*/
        return;
    }
    switch (symbolPtr->type) {
      case SYM_INT:
      case SYM_UNSIGNED:
      case SYM_BOOL:
      case SYM_HYPER:
      case SYM_UNSIGNEDHYPER:
      case SYM_FLOAT:
      case SYM_DOUBLE:
      case SYM_QUADRUPLE:
      case SYM_VOID:
      case SYM_CONST:
        ckfree((char *)symbolPtr);
        break;
      case SYM_ARRAY:
      case SYM_POINTER:
        FreeSymbol(symbolPtr->referenceValue);
        ckfree((char *)symbolPtr);
        break;
      case SYM_TYPE:
        ckfree((char *)symbolPtr->identValue);
        ckfree((char *)symbolPtr);
        break;
      case SYM_ENUM:
        Tcl_DeleteHashTable(symbolPtr->hashTablePtr);
        ckfree((char *)symbolPtr->hashTablePtr);
        ckfree((char *)symbolPtr);
        break;
      case SYM_STRUCT:
        FreeSymbolTable(symbolPtr->hashTablePtr);
        ckfree((char *)symbolPtr->hashTablePtr);
        ckfree((char *)symbolPtr);
        break;
      case SYM_STRING:
      case SYM_OPAQUE:
        ckfree((char *)symbolPtr);
        break;
      case SYM_UNION:
        FreeSymbol(symbolPtr->referenceValue);
        ckfree((char *)symbolPtr->identValue);
        if (symbolPtr->defaultValue) {
            FreeCase((CaseData *)symbolPtr->defaultValue);
        }
        /* 
         * Free all symbols representing the case values. 
         */
        FreeCaseTable(symbolPtr->hashTablePtr);
        ckfree((char *)symbolPtr->hashTablePtr);
        ckfree((char *)symbolPtr);
        break;
      case SYM_PROGRAM:
      case SYM_VERSION:
      case SYM_PROCEDURE:
        FreeSymbolTable(symbolPtr->hashTablePtr);
        ckfree((char *)symbolPtr->hashTablePtr);
        ckfree((char *)symbolPtr);
        break;

      default:
        /*
          printf("internal error: cannot free symbol type %d\n", 
               symbolPtr->type);
               */
        break;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * FreeCaseTable --
 *
 *    Deallocate a hash table containing case structures.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static void 
FreeCaseTable(tablePtr) 
     Tcl_HashTable *tablePtr;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    
    entryPtr = Tcl_FirstHashEntry(tablePtr, &searchInfo);
    while (entryPtr) {
        FreeCase((CaseData *)Tcl_GetHashValue(entryPtr));
        Tcl_SetHashValue(entryPtr, (ClientData)NULL);
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    Tcl_DeleteHashTable(tablePtr);
}

/*
 *----------------------------------------------------------------------
 *
 * FreeCase --
 *
 *    Deallocate a case structure.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static void 
FreeCase(caseDataPtr)
     CaseData *caseDataPtr;
{
    FreeSymbol(caseDataPtr->symbolPtr);
    ckfree((char *)caseDataPtr);
} 

/*
 *----------------------------------------------------------------------
 *
 * FreeSymbolTable --
 *
 *    Free the contents of a hash table (exclusive of the table).
 *    The entries in the table are symbols.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    All memory allocated by symbols is freed.
 *
 *----------------------------------------------------------------------
 */

static void
FreeSymbolTable(tablePtr)
    Tcl_HashTable *tablePtr;
{
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    SymbolData *symbolPtr;

    /*
     * Free all data allocated in the symbol table.
     */
    entryPtr = Tcl_FirstHashEntry(tablePtr, &searchInfo);
    while (entryPtr) {
        symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        FreeSymbol(symbolPtr);
        Tcl_SetHashValue(entryPtr, (ClientData)NULL);
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    } 
    Tcl_DeleteHashTable(tablePtr);
}

/*
 *----------------------------------------------------------------------
 *
 * PrintSymbol --
 *
 *    Add a textual description of a symbol to the tapePtr->interpreter result.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Modifies tapePtr->interpreter result.
 *
 *----------------------------------------------------------------------
 */

static int
PrintSymbol(tapePtr, dsPtr, symbolPtr, name)
     Txrx_Tape *tapePtr;
     Tcl_DString *dsPtr;
     SymbolData *symbolPtr;
     char *name;
{
    static char *types[] = {
        "CONST", "ENUM", "STRUCT", "UNION", "PROGRAM", "VERSION", "PROC",
        "STRING", "OPAQUE", "INT", "UNSIGNED", "BOOL", "HYPER", "UHYPER",
        "FLOAT", "DOUBLE", "QUAD", "VOID", "ARRAY", "POINTER", "TYPE"
    };
    char buf[256];
    int argc, i;
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    CaseData *caseDataPtr;

    if (name == NULL) {
        name = "<NULL>";
    }
    if (name[0] == '_' && name[1] == '_') {
        return TCL_OK;;
    }
    if (symbolPtr == NULL) {
        return TCL_ERROR;
    }
    Tcl_DStringStartSublist(dsPtr);
    Tcl_DStringAppendElement(dsPtr, types[symbolPtr->type]);
    switch(symbolPtr->type) {
      case SYM_CONST:
        Tcl_DStringAppendElement(dsPtr, name);
        sprintf(buf, "%ld", symbolPtr->constValue);
        Tcl_DStringAppendElement(dsPtr, buf);
        break;
      case SYM_POINTER:
        Tcl_DStringAppendElement(dsPtr, name);
        PrintSymbol(tapePtr, dsPtr, symbolPtr->referenceValue, "");
        break;
      case SYM_ARRAY:
      case SYM_STRING:
      case SYM_OPAQUE:
        Tcl_DStringAppendElement(dsPtr, name);
        sprintf(buf, "%ld", symbolPtr->constValue);
        Tcl_DStringAppendElement(dsPtr, buf);
        Tcl_DStringAppendElement(dsPtr,
                          (symbolPtr->identValue == NULL)?"fixed":"variable");
        PrintSymbol(tapePtr, dsPtr, symbolPtr->referenceValue, "");
        break;
      case SYM_TYPE:
        Tcl_DStringAppendElement(dsPtr, symbolPtr->identValue);
        break;
      case SYM_ENUM:
        /* 
         * Traverse the hash table and gather name/value pairs.
         */
        Tcl_DStringAppendElement(dsPtr, name);
        entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
        while (entryPtr) {
            long value = (long)Tcl_GetHashValue(entryPtr);
            Tcl_DStringAppendElement(dsPtr, 
                                     Tcl_GetHashKey(symbolPtr->hashTablePtr, 
                                                    entryPtr));
            sprintf(buf, "%ld", value);
            Tcl_DStringAppendElement(dsPtr, buf);
            entryPtr = Tcl_NextHashEntry(&searchInfo);
        }
        break;
      case SYM_STRUCT:
        /*
         * Traverse hash table and recursively print the 
         * component symbols.
         */
        Tcl_DStringAppendElement(dsPtr, name);
        argc = (int)symbolPtr->constValue;
        for (i = 1; i <= argc; i ++) {
            sprintf(buf, "%d", i);
            entryPtr = Tcl_FindHashEntry(symbolPtr->hashTablePtr, buf);
            if (entryPtr == NULL) {
                Tcl_DStringAppendElement(dsPtr, 
                                         "error - no entry for struct field");
            }
            PrintSymbol(tapePtr, dsPtr, 
                        (SymbolData *)Tcl_GetHashValue(entryPtr), buf);
        }
        break;
      case SYM_UNION:
        /*
         * Print discriminant.
         */
        Tcl_DStringAppendElement(dsPtr, name);
        PrintSymbol(tapePtr, dsPtr, 
                    symbolPtr->referenceValue, 
                    symbolPtr->identValue);
        entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
        while (entryPtr) {
            caseDataPtr = (CaseData *)Tcl_GetHashValue(entryPtr);
            sprintf(buf, "%s", 
                    Tcl_GetHashKey(symbolPtr->hashTablePtr, entryPtr));
            Tcl_DStringAppendElement(dsPtr, buf);
            PrintSymbol(tapePtr, dsPtr,
                        caseDataPtr->symbolPtr, caseDataPtr->symbolNamePtr);
            entryPtr = Tcl_NextHashEntry(&searchInfo);
        }
        /*
         * Print the default case value.
         */
        if (symbolPtr->defaultValue) {
            caseDataPtr = (CaseData *)symbolPtr->defaultValue;
            Tcl_DStringAppendElement(dsPtr, "default ");
            PrintSymbol(tapePtr, dsPtr,
                        caseDataPtr->symbolPtr, caseDataPtr->symbolNamePtr);
        }
        
        break;
      case SYM_PROGRAM:
      case SYM_VERSION:
        Tcl_DStringAppendElement(dsPtr, name);
        sprintf(buf, "%ld", symbolPtr->constValue);
        Tcl_DStringAppendElement(dsPtr, buf);
        entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
        while (entryPtr) {
            PrintSymbol(tapePtr, dsPtr,
                        (SymbolData *)Tcl_GetHashValue(entryPtr),
                        Tcl_GetHashKey(symbolPtr->hashTablePtr, entryPtr));
            entryPtr = Tcl_NextHashEntry(&searchInfo);
        }
        break;
      case SYM_PROCEDURE:
        Tcl_DStringAppendElement(dsPtr, name);
        sprintf(buf, "%ld", symbolPtr->constValue);
        Tcl_DStringAppendElement(dsPtr, buf);
        argc = (int)symbolPtr->identValue;
        for (i = 0; i <= argc; i ++) {
            sprintf(buf, "%d", i);
            entryPtr = Tcl_FindHashEntry(symbolPtr->hashTablePtr, buf);
            if (entryPtr == NULL) {
                Tcl_DStringAppendElement(dsPtr, 
					 "no entry for procedure field");
            }
            PrintSymbol(tapePtr, dsPtr,
                        (SymbolData *)Tcl_GetHashValue(entryPtr),
                        Tcl_GetHashKey(symbolPtr->hashTablePtr, entryPtr));
        }
        break;
      case SYM_INT:
      case SYM_UNSIGNED:
      case SYM_BOOL:
      case SYM_HYPER:
      case SYM_UNSIGNEDHYPER:
      case SYM_FLOAT:
      case SYM_DOUBLE:
      case SYM_QUADRUPLE:
      case SYM_VOID:
        break;
      default:
        sprintf(buf, "%d", symbolPtr->type);
        Tcl_DStringAppendElement(dsPtr, buf);
        break;
    }
    Tcl_DStringEndSublist(dsPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * InitSymbol --
 *
 *    Initialize with default values a symbol.
 *
 * Results:
 *    Pointer to allocated symbol.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static SymbolData*
InitSymbol()
{
    SymbolData *symbolPtr = (SymbolData *)ckalloc(sizeof(SymbolData));
    symbolPtr->constValue = 0l;
    symbolPtr->identValue = NULL;
    symbolPtr->referenceValue = NULL;
    symbolPtr->defaultValue = NULL;
    symbolPtr->hashTablePtr = NULL;
    symbolPtr->type = SYM_VOID;
    symbolPtr->codeAddress = 0;

    return symbolPtr; 
}

/*
 *----------------------------------------------------------------------
 *
 * InitCase --
 *
 *    Initialize a case alternative.
 *
 * Results:
 *    Pointer to case description.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static CaseData*
InitCase(symbolPtr, label, name)
     SymbolData *symbolPtr;
     Token *label;
     char *name;
{
    CaseData *caseDataPtr = (CaseData *)ckalloc(sizeof(CaseData));

    caseDataPtr->symbolPtr = symbolPtr;
    strcpy(caseDataPtr->symbolNamePtr, name);
    caseDataPtr->label = *label;
    
    return caseDataPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * <procName> --
 *
 *    <description>
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

void
Routine()
{
}
